################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "devops_core/util"
require "ezcrypto" # MIT license ezcrypto-0.7.2.gem

################################################################################
# Class Definition                                                             #
################################################################################
module DevopsCore
  module CryptoUtil

    # Create AES 128-bit Key
    # => 1. Create a strong alpha-numeric password longer then 8-bytes
    # => 2. Create a Salt or two-form hash
    DEFAULT_PASSWORD = "E88A3EA1-C3CE-BED1-D7DD-D00D0A54720F"
    DEFAULT_SALT = "66D6DF2E-68AA-24F0-3506-6D942F4FF0FB"
    @key = EzCrypto::Key.with_password(DEFAULT_PASSWORD, DEFAULT_SALT)
    
    # The format of passwords that need to be decrypted.  Like this: decrypt('something') or decrypt("something").
    # A string: whitespace, "decrypt(", whitespace, quote, string, same-quote, whitespace, ")", whitespace
    DECRYPT_REGEXP = /\A\s*decrypt\(\s*(['"])(.*)\1\s*\)\s*\z/
    
    # Logger setup
    @logger = Logger.new(STDOUT)
    
    # This method will generate a new keyfile using a hash salt and a password.  This creates an AES 128-bit 
    # symmetric key.  The keyfile should be stored in a private component (weaker security) or a credential keystore
    # (better security).
    #
    # == Params
    #
    # === Required
    # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
    #
    # === Optional
    # [password (String)] A password that's used the generate the key.  If this is nil, DEFAULT_PASSWORD will be used.
    # [salt (String)] A salt value for the hash.  You can consider this another password that's used to generate the key.
    #    If this is nil, DEFAULT_SALT will be used.
    #
    # == Returns
    # [String] The path of the keyfile, or nil if it wasn't created because a parameter was missing.  
    #
  def CryptoUtil.generate_key_file(key_file_name, password, salt)

    if password.nil?
      @logger.warn "using default password"
      password = DEFAULT_PASSWORD
    end 
    if salt.nil?
      @logger.warn "using default hash salt value"
      salt = DEFAULT_SALT
    end 
    
    if key_file_name.nil?
      @logger.error "key_file_name is required"
      return nil
    end 
    
    # Create AES 128-bit Key
    current_key = EzCrypto::Key.with_password(password, salt)
    current_key.store(key_file_name)
    
    return key_file_name
  end # end def CryptoUtil.generate_key_file(password, salt, key_file_name)

    # This method will generate a key and encrypt a value.
    # It will generate a new keyfile if the file does not exist and key_file_name is provided.
    # If you see this message, use encrypt64/decrypt64 instead: Input length (with padding) not multiple of 16 bytes
    #
    # == Params
    #
    # === Required
    # [value (String)] The value to encrypt.
    #
    # === Optional
    # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
    #
    # == Returns
    # [String] The encrypted value.  
    #
	def CryptoUtil.encrypt(value, key_file_name = nil)
		current_key = @key
		final_value = value

		if key_file_name and File.exists?(key_file_name)
			current_key = EzCrypto::Key.load(key_file_name)
		elsif key_file_name
			current_key.store(key_file_name)
		end
		
		final_value = current_key.encrypt value
		return final_value
	end # end def CryptoUtil.encrypt(value, key_file_name)
	
	  # This method will decrypt a value using a keyfile.
    # It will use an internal key if the file does not exist.
    # If you see this message, use encrypt64/decrypt64 instead: Input length (with padding) not multiple of 16 bytes
    #
    # == Params
    #
    # === Required
    # [value (String)] The value to decrypt.
    #
    # === Optional
    # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
    #
    # == Returns
    # [String] The decrypted value.  
    #
	def CryptoUtil.decrypt(value, key_file_name = nil)
		current_key = @key
		final_value = value

		if key_file_name and File.exists?(key_file_name)
			current_key = EzCrypto::Key.load(key_file_name)
		else
			current_key = @key
		end
		
		final_value = current_key.decrypt value
		return final_value
	end # end def CryptoUtil.decrypt(value, key_file_name)

  # This method will generate a key and encrypt a value, then convert it to base 64 notation.
  # It will generate a new keyfile if the file does not exist and key_file_name is provided.
  #
  # == Params
  #
  # === Required
  # [value (String)] The value to encrypt.
  #
  # === Optional
  # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
  #
  # == Returns
  # [String] The encrypted value.  
  #
  def CryptoUtil.encrypt64(value, key_file_name = nil)
    current_key = @key
    final_value = value

    if key_file_name and File.exists?(key_file_name)
      current_key = EzCrypto::Key.load(key_file_name)
    elsif key_file_name
      current_key.store(key_file_name)
    end
  
    final_value = current_key.encrypt64 value
    return final_value.strip! # for some reason this was appending a \n character
  end # end def CryptoUtil.encrypt64(value, key_file_name)

  # This method will decrypt a value in base 64 notation using a keyfile.
  # It will use an internal key if the file does not exist.
  #
  # == Params
  #
  # === Required
  # [value (String)] The value to decrypt.
  #
  # === Optional
  # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
  #
  # == Returns
  # [String] The decrypted value.  
  #
  def CryptoUtil.decrypt64(value, key_file_name = nil)
    current_key = @key
    final_value = value

    if key_file_name and File.exists?(key_file_name)
      current_key = EzCrypto::Key.load(key_file_name)
     else
      current_key = @key
     end
  
     final_value = current_key.decrypt64 value
     return final_value
  end # end def CryptoUtil.decrypt64(value, key_file_name)
  
  # This method will decrypt a value in base 64 notation using a keyfile,
  # if the value is an expression in the form decrypt('base64text').
  # If the value is not in that form, it will be returned unchanged.
  # If the keyfile does not exist, it will use an internal key.
  #
  # == Params
  #
  # === Required
  # [propname (String)] The name of the property to be decrypted; used to print a warning 
  #    if the property is not encrypted.
  # [possibly_encrypted_string (String)] The value to be decrypted.  May be either 
  #    an expression like decrypt('base64text') OR a plain string.
  #
  # === Optional
  # [key_file_name (String)] The path of the keyfile to use.  This is a yaml file generated by EzCrypto::Key#store.
  #
  # == Returns
  # [String] The decrypted value, if it was an encrypted string; otherwise, returns possibly_encrypted_string.
  #
  def CryptoUtil.decrypt_if_needed(propname, possibly_encrypted_string, key_file_name = nil)
    return nil if possibly_encrypted_string.nil?

    if key_file_name.nil?
      key_file_name = ENV['KEY_FILE']
    end
    
    match = DECRYPT_REGEXP.match(possibly_encrypted_string)
    if match.nil?
      @logger.warn "#{propname} is stored as plaintext"
      return possibly_encrypted_string
    end 

    encoded = Base64.decode64(match[2])
    begin
      final_value = DevopsCore::CryptoUtil.decrypt(encoded, key_file_name)
    rescue OpenSSL::Cipher::CipherError => ce
      @logger.error "Can't process #{match[2].inspect} using keyfile #{key_file_name.inspect}"
      raise # raises ce again
    end
    return final_value
  end # end CryptoUtil.decrypt_if_needed(propname, possibly_encrypted_string, key_file_name = nil)
  
  # This method convert an expression in the form decrypt('base64text') into 
  # just the base64text.
  # If the value is not in that form, it will return nil.
  #
  # == Params
  #
  # === Required
  # [expr_string (String)] The expression to parse.
  #
  # == Returns
  # [String] The base64 value within the expression.
  #
  def CryptoUtil.base64_from_expression(expr_string)
    match = DECRYPT_REGEXP.match(expr_string)
    return match[2]
  end # end CryptoUtil.base64_from_expression(expr_string)

end # end CryptoUtil
  
end # end DevopsCore
