################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require "erb" # Ruby templating engine
require "json" # JSON parsing
require 'open-uri'
require 'openssl'

################################################################################
# Class Definition                                                             #
################################################################################
module DevopsCore
  module Util

    ################################################################################
    # JSON Methods                                                                 #
    ################################################################################
    
    # A method that till take a JSON string or file and convert its contents to 
    # a Hash object.
    #
    # == Params
    #
    # === Required
    # [json (String|URI String)]  A JSON formatted string or URI to a local or remote
    #                             JSON file.
    #
    # == Returns
    # [Hash]  A hashmap object representing the key-value pairs of the original JSON content.
    #
    # == Exceptions
    # [JSON::ParserError] This exception will occur when the content in the string or file
    #                     passed in as an argument is not correctly formatter as JSON
    #
    def Util.create_hash_from_json(json)
      begin

        json_obj = ""
        json_obj = JSON.parse(json)
    
        return json_obj
    
      rescue
        
        json_content = ""
    
        if File.file?(json) or !(json =~ URI::regexp).nil?
          json_content = open(json) { |f| f.read }
        end
    
        return JSON.parse(json_content)

      end
      
    end
    
    # Determine if a string is JSON formatted
    #
    # == Params
    #
    # === Required
    # [str (String)]  A string to test for JSON formatting
    #
    # == Returns
    # * True if the provided string is valid JSON
    # * False if the provided string is not valid JSON
    #
    def Util.is_json(str)
      json_content = str
      
      if File.file?(str) or !(str =~ URI::regexp).nil?
        json_content = open(str) { |f| f.read }
      end
      
      begin
        JSON.parse(json_content)
        return true
      rescue 
        return false
      end
      
    end
  
    ################################################################################
    # Template Methods                                                             #
    ################################################################################
    
    # Evaluate an ERB template by combining the template string to the values declared
    # in the context object.
    #
    # == Params
    #
    # === Required
    # [template (String)]      An ERB formatted template string.
    # [context (Hash|Object)]  If a hashmap is provided, the key-value pairs will be converted into an object's 
    #                          instance variables and accessed in a template via the "@key" syntax. If an object
    #                          is provided as the value, the ERB engine will use the instance's attributes
    #                          when evaluating the template string.
    #
    # == Returns
    # [String]  A complete template with its original variables filled in.  
    #
    def Util.evaluate_template(template, context)
    
      binding = context

      if  context.is_a?(Hash)
        bh = BindingHelper.new(context)
        binding = bh.get_binding
      end

      begin
        template_obj = ERB.new(template)
        return template_obj.result(binding)
      rescue Object => e
        puts e.inspect
        puts e.backtrace.join("\n")
        raise StandardError, "The template failed to be evaluated!"
      end
  
    end

    # This internal class helps with generating an object
    # that has all of the Environment bindings attached to
    # class attributes. This is done because the ERB template
    # API requires a Binding object to determine the context
    # in which it looks for the values of the template variables
    class BindingHelper
      def initialize(hash_obj)
        hash_obj.each do |k, v|
          self.instance_variable_set("@#{k}".to_sym,v)
        end
      end
      
      def get_binding
        return binding()
      end
    end
  
  end
end
