################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

################################################################################
# Dependencies                                                                 #
################################################################################
require "logger"
require "fileutils"
require "uri"
require "net/http"
require "devops_core/crypto_util"

module DevopsLibrary  
    class Download
      
      @@logger = Logger.new(STDOUT)
      MAX_REDIRECTS =5
        # Download a file from a server.
        #
        # Params
        # ======
        #   Required
        #   --------
        #     url (string): The url of the file to download.
        #     path (string): The location on the local file system to save the file to 
        #     user (string): User Id 
        #     password (string): Password for the user
        #     key_file (string): Key file to decrcypt the password if needed
      
        def self.download (url, path, user, password, key_file=nil)  
           raise ArgumentError, "url is required." if url.nil?
           raise ArgumentError, "path is required." if path.nil?
           raise ArgumentError, "path should be a directory" if File.exists?(path) && !File.directory?(path)
                     
           filename = File.basename(url)
           file = "#{path}/#{filename}"
           @@logger.info "Downloading URL : #{url} to file"
           #Decrypt the password if needed
           password = DevopsCore::CryptoUtil.decrypt_if_needed("password", password, key_file)
           #Create the directory   
           Dir.mkdir(path) if !File.exists?(path) 
           #Download the file
           internal_download(url, file , user, password)  
           return file    
        end
        
        private 
        
        ##############################################################################
        # Private Methods                                                            #
        ############################################################################## 
        def self.internal_download(url, file, user, password, redirects=0) 
           raise "Reached maximum number of redirects" if (redirects > MAX_REDIRECTS)
              
           uri = URI.parse(url)
           http = Net::HTTP.new(uri.host, uri.port) 
            
           if uri.scheme == "https"
              http.use_ssl = true
              http.verify_mode = OpenSSL::SSL::VERIFY_NONE
           end 
           request = Net::HTTP::Get.new(uri.path) 
           if !uri.user.nil?
              user = uri.user 
           end
            
           if !uri.password.nil?
             password = uri.password
           end
             
           request.basic_auth user, password if (!user.nil? && user.length >0)
           
           http.request request do |response|
              if response.kind_of?(Net::HTTPRedirection)      
                  url = redirect_url(response) 
                  @@logger.info "Redirecting to #{url}"
                  internal_download(url, file, user, password, redirects+1) 
              else
                  f = open(file, "w")
                  begin
                      response.read_body do |chunk|
                          f.write(chunk)
                      end
                  ensure
                      f.close()
                  end
              end #end of if
            end #end of http
        end # end of internal download
        
       def self.redirect_url(response)
          if response['location'].nil?
            response.body.match(/<a href=\"([^>]+)\">/i)[1]
          else
            response['location']
          end
        end
    end  
end