################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

module DevopsLibrary
  module Providers
    class LibrarySystemProvider
      
      TEST_SUCCESS="Success"
      TEST_FAIL="Fail"
      
      # Creates a RAM asset and optionally publishes an artifact(s) to it.
      #  
      # Parameters
      # ==========
      #   package_id      : GUID of the asset
      #   package_version : Version of the asset
      #   metadata        : Additional metadata that can be used to create the asset
      #        build_url       : URL of the build
      #        name            : Application name
      #        description     : description of the asset to be created. 
      #                          If description is not passed, name would be set as description
      #        community_name  : Name of the community
      #        asset_type      : Asset Type of the asset
      #        attributes      : array of name / value
      #                          attribute_name : name of the attribute
      #                          attribute_value : value of the attribute
      #        classifications : classifications for the asset
      #
      #   packages : Hash containing the files to be published to the asset
      #              fully qualified file name => file name
      #   optimize : Flag signifying whether or not asset creation is optimized.  If the
      #              value is true then the asset is only created if the artifacts to be
      #              included in this new asset are different than the artifacts included
      #              in an existing asset (if one exists).  If the value is false then
      #              the asset is always created and the artifacts published to it.  The
      #              default value is true.
      #   
      # Output
      # ======
      #    Package : DevopsLibrary::Providers::Package 
      #
      def create_package(package_id, package_version, metadata={}, packages=nil, optimize=true)
        raise NotImplementedError
      end
  
      # Adds file to the package
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #
      #   file : file to be published to the package
      # Output
      # ======
      #    url of the file in the package
      #
     def publish_package(package, file)
        raise NotImplementedError
      end
      
      # Downloads the package to the folder and returns the path of the file
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       folder : Directory to which the artifact should be downloaded
      #       filename : Save As file name
      #
      def fetch_package(package_id, package_version, folder, filename=nil)
          raise NotImplementedError
      end 
      
      # Fetches the artifacts matching the regular expression from the asset and download it to the folder
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       filename : name or regular expression matching the artifact
      #       folder : Directory to which the artifact should be downloaded
      #
      #
      #   Return
      #       Absolute file path
      #
      def fetch_package_files(package_id, package_version, filename, folder) 
        raise NotImplementedError
      end 
      
      
      # Fetches the file from the package and downloads it to the output_folder
      # Parameters
      # ===========
      #   Required
      #       package_id : ID of the package
      #       package_version : Version of the package
      #       filename : name of the file in the package
      #       folder : Directory to which the file should be downloaded
      #
      #   Optional     
      #       metadata : additional parameters. Not needed for ram library provider
      #
      def fetch_package_file(package_id, package_version, filename, folder)
        raise NotImplementedError
      end 
      
      # Generates the deps file for the given asset
      # Parameters
      # ===========
      #   Required
      #       package_id : GUID of the asset
      #       package_version : Version of the asset
      #       
      #   Return
      #       url of the Deps file
      #
      def publish_dependency_file(package)
        raise NotImplementedError
      end
      
      # Resolves the package with the highest version of the package
      # 
      # Parameters
      # ===========
      #   Required
      #       package_id : ID of the package
      #       package_version : Version of the asset or version range
      #       metadata : Additional query parameters
      #       
      #   
      # Output
      # ======
      #    package (structure)
      #          id : ID of the library package
      #          version : Resolved version of the package
      #          metadata : Provider specific metadata describing the additional information for the package
      #          url : URL of the package
      #           
      #
      def resolve_package(package_id, package_version, metadata={})
        raise NotImplementedError
      end 
      
      def add_attribute(package, attribute_name, attribute_value, attribute_url=nil)
        raise NotImplementedError
      end 
      
      # Returns the related assets as an array
      # Parameters
      # ===========
      #   Required
      #       package_id : id of the package
      #       package_version : Version of the package
      #       related_types: Relationship between the packages to download (default={'dependency', 'Infrastructure'})
      #       recursive : true or false (default= true)
      #          true : Recursive fetches the parents 
      #       
      #   Return
      #       Array of the related asset package structure
      #
      def get_related(package_id, package_version, related_types=nil, recursive=true)
         raise NotImplementedError
      end 
      
      # Returns the package as json
      # Parameters
      # ===========
      #   Required
      #       package_id : id of the package
      #       package_version : Version of the package
      #   Return
      #       package metadata as json
      #
      def get_package(package_id, package_version)
         raise NotImplementedError
      end  
       
      # Updates the package from the metadata
      # Parameters
      # ===========
      #   Required
      #       package_id : id of the package
      #       package_version : Version of the package
      #       metadata : asset metadata
      #
      def update_package(package_id, package_version, metadata)
         raise NotImplementedError
      end
      
      # Get URL of the package / file
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    file : artifact name
      #     
      def get_url(package_id, package_version, filename=nil)
        raise NotImplementedError
      end
      
      # Creates a relationship between two packages
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    related_package_id : ID of the related package
      #    related_package_version : Version of the related package
      #    relationship : relationship name
      #   
      def relate_package(package_id, package_version, related_package_id, related_package_version, relationship)
        raise NotImplementedError
      end
      
      # classifies the package with the classification
      #    package_id : ID of the package
      #    package_version : Version of the package
      #    category : category path
      #    recategorize : true, if sibling is exclusive and should be removed before setting the category
      #                 : false, if sibling is exclusive and attempt to categorize the asset will
      #                 : throw exception
      #                 : default is true
      def categorize_package(package_id, package_version, classification, recategorize=true)
        raise NotImplementedError
      end
      
      # Queries server for the list of packages
      #    query : search query
      #    sort : sort criteria for the search
      #    ascending : true or false. Default is ascending
      #
      # Return
      #     packages as json
      def query_packages(query, sort=nil, ascending=true)
        raise NotImplementedError
      end
      
      # Publishes the artifact content link to the asset
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #
      #   url : url to be published as artifact
      #   name : [Optional] If name is nil, name will be the last segment of the url 
      # Output
      # ======
      #    url of the file in the package
      #
      def publish_link(package, url, name=nil)
        raise NotImplementedError
      end
      
      # Sets the status of the package as deployed for the build environment. The environment is 
      # the environment set by the build
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #

      def mark_deployed(package)
        raise NotImplementedError
      end
      
      # Sets the status of the package as test successful or test failed
      # Parameters
      # ==========
      #   package (structure)  
      #       id : ID of the package
      #       version : Version of the package
      #       metadata : additional metadata describing the package
      #       url : url of the package
      #
      #   test_status : DevopsLibrary::Providers::LibrarySystemProvider:: TEST_SUCCESS or DevopsLibrary::Providers::LibrarySystemProvider::TEST_FAIL
      #
      #
      def mark_tested(package, test_status)
        raise NotImplementedError
      end
      
    end
  end  
end # end module DevopsLibrary

  
