require "logger" # Ruby Logging

require 'weaver/core/logging'   
require 'weaver/cli/weaver_cli_plugin'
require 'weaver/dsl/weaver_parser'
require 'weaver/icct/icct_client'
require 'weaver/icct/icct_request_generator'

module Weaver
  module Cli
    module Plugins

      # ICCT plugin for the Weaver CLI
      #
      # ICCT is the IBM Image Construction and Composition Tool
      #
      # Syntax: <tt>weaver icct [task] [options]</tt>
      #
      # Author:: Michael Kalantar
      #
      class IcctCliPlugin < Weaver::Cli::WeaverCliPlugin

        namespace :icct
        namespace_description "ICCT tasks"

        desc "create_bundle", "Create an ICCT composable software bundle."
        method_option :recipe_url, :type => :string, :required => false, :desc => "URL of cookbook containing recipe."
        method_option :recipe, :type => :string, :required => false, :desc => "Recipe to covert to a bundle."
        method_option :chef_path, :type => :string, :required => false, :desc => "Location of Chef cookbooks."
        method_option :role, :type => :string, :required => false, :desc => "Chef role to covert to a bundle."
        method_option :host, :type => :string, :required => false, :desc => "URL of ICCT instance."
        method_option :user, :type => :string, :required => false, :desc => "User of specified ICCT instance."
        method_option :pwd, :type => :string, :required => false, :desc => "Password for specified user of specified ICCT instance."
        method_option :automation, :type => :string, :required => false, :desc => "Automation model."
        
        def create_bundle(name, uuid, version, description = nil)
          setup_logging
          collect_properties

          # This is a bug in Thor and JRuby so we have to handle that manually
          raise Thor::Error.new(%Q{"create_bundle" was called incorrectly. name required.}) unless name
          raise Thor::Error.new(%Q{"create_bundle" was called incorrectly. uuid required.}) unless uuid
          raise Thor::Error.new(%Q{"create_bundle" was called incorrectly. version requried.}) unless version

          chef_path, role_file = identify_recipe()
          @@logger.debug("Processing '#{chef_path}/roles/#{role_file}' to create bundle '#{name}' with UUID '#{uuid}_#{version}'")
          
          # identify parameters for connection for ICCT
          host, username, password = process_icct_connection_info()            
          @@logger.debug("Identified ICCT connection details as: #{username}@#{host}:#{password}")
          client_config = {
              :log_level => Logger::INFO,
              :log_file => STDOUT
          }
          client = Weaver::Icct::IcctClient.new(host, username, password, client_config)
          client.ssl_verify_mode=OpenSSL::SSL::VERIFY_NONE
          
          @@logger.debug("chef automation file is: #{options[:automation]}")
          
          if options[:automation]
             model = Weaver::Dsl::WeaverParser.new().parse(File.new(options[:automation]))
          else
             inspector = Weaver::Chef::ChefNodeInspector.new(chef_path.split(File::PATH_SEPARATOR)) do |level, msg|
               report_message("Importing", level, msg)
             end # end do
             model = inspector.model_role(role_file, ".")
          end # end if options[:automation]
          
          # identify/create files to include in run/config operation for bundle
          cookbook_archive = archive_cookbooks(chef_path)
          run_script = create_run_script(model)
          @@logger.debug("Include in bundle: '#{cookbook_archive}' and '#{run_script}'")
          posted_files = client.add_files({
                "cookbooks" => cookbooks = File.open(cookbook_archive, "rb"),
                "execute" => run = File.open(run_script, "rb")
                   })
          
          # create bundle
          generator = Weaver::Icct::IcctRequestGenerator.new({'name' =>name, 'symbolicName' => uuid, 'version' => version, 'description' => description})
          new_bundle = client.create_bundle(generator.to_json);
          
          # update bundle
          bundle = Weaver::Icct::IcctRequestGenerator.new(new_bundle);
          # don't know the version or name of the software; use bundle name, version
          bundle.add_product(name, version)
          bundle.add_product_requirement("chef-solo", "0.8.1+")
          bundle.add_operation(Weaver::Icct::IcctRequestGenerator::IMAGE_BUILD, role_file, model, posted_files)
          bundle.add_operation(Weaver::Icct::IcctRequestGenerator::IMAGE_DEPLOY, role_file, model, posted_files)
          client.update_bundle(bundle.id, bundle.to_json);

          return bundle.id
        end # end create_bundle method
        
        no_tasks do
            def identify_recipe()
               @@logger.debug("Identifying recipe to convert")
               # 'chef_path' is a path to a directory containing cookbooks, roles and nodes
               # same as 'file_cache_path' in a chef-solo config file
               chef_path = options[:chef_path] || ENV['CHEF_SETUP_PATHS']
               @@logger.debug("Chef setup claimed to be in: #{chef_path}")
               validate_path(chef_path) unless chef_path.nil?
            
               # download cookbook if needed
               url = options[:recipe_url]
               @@logger.debug("Need to download cookbook '#{url}'") if url
               chef_path = fetch_cookbook(url, chef_path) if url
               say_status("Error", "Parameter 'path' (or environment variable 'CHEF_SETUP_PATHS') not set.", :red) && exit(1) unless chef_path
               @@logger.debug("Chef setup confirmed to be in: #{chef_path}")
               
               # chef_path should be defined and cookbook should be in chefpath/cookbooks
               cookbook, recipe = process_recipe_info()
               @@logger.debug("Will convert recipe: #{cookbook}::#{recipe}") unless cookbook.nil?
               
               role = options[:role] || (create_role(cookbook, recipe, chef_path) unless cookbook.nil?)
               @@logger.debug("Converting role: '#{role}'") unless role.nil?
               say_status("Error", "No recipe identified for conversion to bundle", :red) && exit(1) unless role
               
               return chef_path, role                
            end # end process_recipe_info method
            
            def fetch_cookbook(url, path)
               @@logger.debug("Fetching cookbook '#{url}' to '#{path}'")
               # create_chef_files if path.nil?
               return path
            end # end fetch_cookbook method
            
            def validate_path(path)
               @@logger.debug("validating path '#{path}'")
               unless path.nil?
                  path.split(File::PATH_SEPARATOR).each do |dirname|
                     @@logger.debug("validating sub element '#{dirname}'")
                     unless File.directory?(dirname)
                        say_status("Error", "Directory #{dirname} does not exist", :red) && exit(1)
                     else
                        say_status("Error", "#{dirname} is not a Chef directory (ie, does not contain 'cookbooks'; or 'roles' subdirectories.", :red) && exit(1) unless File.directory?(File.expand_path('cookbooks', dirname)) || File.directory?(File.expand_path('roles', dirname))
                     end # end unless File.directory?(dirname)
                  end # each do
               end # end unless path.nil?
            end # end validate_path method
            
            def process_recipe_info()
               @@logger.debug("Looking for recipe to convert")
               recipe = options[:recipe]
               unless recipe.nil?
                  cr = recipe.split("::")
                  cookbook = cr[0];
                  if (1 == cr.length) 
                     recipe = "default"
                  else 
                     recipe = cr[1]
                  end # if (1 == cr.length)
               end # end unless recipe.nil?
               
               return cookbook, recipe
            end # end process_recipe_info method
            
            def create_role(cookbook, recipe, chef_path)
               @@logger.debug("Creating role for '#{cookbook}::#{recipe} in #{chef_path}'")
               role = cookbook + "_" + recipe
               role_filepath = "#{chef_path}/roles/" + role + "_role.rb"
               file = File.open(role_filepath, 'w')
               file.write(%Q{name "#{cookbook}"\n})
               file.write(%Q{description "Generated role for deploying #{cookbook}::#{recipe}"\n})
               file.write(%Q{run_list "recipe[#{cookbook}::#{recipe}]"})
               return role
            end # end create_role method

            def process_icct_connection_info()
              host = options[:host] || ENV['ICCT_HOSTNAME']
              host = "https://" + host if (host =~ /^https:\/\//).nil?
              say_status("Error", "Argument 'host' or environment variable 'ICCT_HOSTNAME' not set.", :red) && exit(1) unless host
              username = options[:user] || ENV['ICCT_USER']
              say_status("Error", "Argument 'user' or environment variable 'ICCT_USER' not set.", :red) && exit(1) unless username
              password = options[:pwd] || ENV['ICCT_PASSWORD']
              say_status("Error", "Argument 'pwd' or environment variable 'IWD_PASSWORD' not set.", :red) && exit(1) unless password
              return host, username, password
            end # end process_icct_connection_info method
            
          # @private
          # :stopdoc:
          # Report a status message running a command
          def report_message(status, level, msg)
            if level == Logger::ERROR || level == Logger::FATAL
              say_status(status, msg, :red)
            elsif level == Logger::WARN
              say_status(status, msg, :yellow)
            else
              say_status(status, msg, :green)
            end
          end
          # :startdoc:
            
           def archive_cookbooks(chef_path)
              cookbook_dir = chef_path
              cookbook_archive = "cookbooks.tgz"
              # cookbook_dir = find_cookbook(chef_path)
              `tar -czf #{cookbook_archive} -C #{cookbook_dir} .`
              return cookbook_archive
           end # end archive_cookbooks method
           
           def create_run_script(model)
              return  File.join(File.dirname(__FILE__), "..", "..", "icct", 'bin', 'chef_bundle_run.sh')
           end # end create_run_script method
            
        end # end no_tasks do
      end # end ChefPlugin class
    end # end Plugins Module
  end # end Cli Module
end # end Weaver Module
