require 'weaver/cli/weaver_cli_plugin'
require 'thor/actions'
require 'weaver/core/constants'
require "devops_core/crypto_util"
require "base64"
require 'fileutils'

module Weaver
  module Cli
    module Plugins

      # Project plugin for the Weaver CLI
      #
      # Syntax: weaver key [task] [options]
      #
      # Author:: Ed Snible
      #
      class KeyCliPlugin <  Weaver::Cli::WeaverCliPlugin
        include Thor::Actions

        # Special non-Thor extension to provide help for the arguments of the commands
        ARG_DEFS = {
          'FILENAME' => 'The name of the new key file',
          'STRING' => 'The string to encrypt, such as a password',
          'BASE64' => 'The string to decrypt',
        }
        
        namespace :key
        namespace_description "Crypto related tasks"

        desc "create FILENAME", "Create a key file"
        method_option :algorithm, :type => :string, :required => false, :desc => "(optional) bf|blowfish|des|des3|aes128|aes192|aes256|rc2|rc4"
        method_option :loc, :type => :string, :aliases => "-l", :desc => "Location of the file."
        # Create an ezcrypto-format keyfile with .key extension
        def create(filename)
          collect_properties
          setup_logging
          
          unless options[:loc].nil?
            filename = File.join(options[:loc], filename)
          end
          
          key = EzCrypto::Key.generate(options)
          
          FileUtils.mkdir_p(File.dirname(filename)) unless File.directory?(File.dirname(filename))
          begin
            key.store(filename)
            say_status("Created", filename, :green)
            say_status("", "You should now 'export KEY_FILE=#{filename}'", :green)
          rescue Errno::EEXIST => ex
            say_status("Problem", ex.message, :red)
            exit(1)
          end
        end

        desc "encrypt STRING", "Encrypt a password string"
        method_option :keyfile, :type => :string, :required => false, :desc => "(optional) key file"
        # Encrypt a string using --keyfile
        def encrypt(str = nil)
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"encrypt" was called incorrectly. Call as "encrypt STRING".}) unless str
          
          collect_properties
          setup_logging

          keyfile = options[:keyfile]
          if keyfile.nil? && !ENV['KEY_FILE'].nil?
            keyfile = ENV['KEY_FILE']
          end
          
          if keyfile.nil?
            say_status("Warning", "Using built-in key, consider specifying --keyfile", :yellow)
          elsif !File.exist?(keyfile)                      
            say_status("Error", "Key file #{keyfile.inspect} does not exist", :red)
            return
          elsif options[:keyfile].nil?
            # If no explicit --keyfile was specified, but there is a KEY_FILE environment variable, let the use know we will use it
            say_status("Using", keyfile, :green)                      
          end
          
          enc = EncryptedValue.from_plaintext(str, keyfile)
          say_status("Use", enc.inspect, :green)                      
        end
        
        desc "decrypt BASE64", "Decrypt the output of 'encrypt'"
        method_option :keyfile, :type => :string, :required => false, :desc => "(optional) key file"
        # Decrypt a String that was previously encrypted with keyfile.
        def decrypt(str = nil)
          # Limitation in Thor and JRuby so we have to missing parameters by assigning defaults and checking for them
          raise Thor::Error.new(%Q{"decrypt" was called incorrectly. Call as "decrypt STRING".}) unless str
          
          collect_properties
          setup_logging
          
          if options[:keyfile].nil? && !ENV['KEY_FILE'].nil?
            say_status("Using", ENV['KEY_FILE'], :green)                      
          end
          
          begin
            encoded = Base64.decode64(str)
            plaintext = DevopsCore::CryptoUtil.decrypt(encoded, options[:keyfile] || ENV['KEY_FILE'])
            say_status("Plaintext", plaintext.inspect, :green)
          rescue Exception => err
            @@logger.debug "Couldn't decode, backtrace=#{err.backtrace}"
            say_status("Problem", err.message, :red)
            say_status("Problem", 'This may indicate a missing --keyfile option', :yellow)
            exist(1)
          end                      
        end
        
      end
    end

  end
end