require 'weaver/core/base'
require 'weaver/core/generation'
require 'weaver/core/late_binding'

module Weaver
  module Core
    # Implements core functionality for an attribute attached to a Weaver model element (AKA a
    # Base or a Weaver entity).  Attributes have types and values or can have a LateBinding
    # as a value which indicates the value is a reference for another value in the model.
    #
    # Author:: Florian Rosenberg
    # Author:: Ed Snible
    class Attribute < Base     
      include Generation

      # After cloning late_binding scopes are stale, this method updates them
      register_reference_freshener :freshen_late_bindings
      
      # The name of the Attribute
      attr_accessor :name

      # The description of the Attribute
      attr_accessor :description

      # Declares whether an attribute is required to have a value set (Default: <code>true</code>).
      # This metadata is similar to the +not_empty+ validator, but only has meaning at +:predeployment+ time.
      attr_accessor :required 
      
      # Specifies the default value of an attribute (Default: nil)
      attr_accessor :default
      
      # Label (String) indicating the short descriptive label (Default: nil)
      attr_accessor :label
      
      # Specifies a grouping of attributes (Default: nil)
      attr_accessor :group
      
      # Specifies whether an attribute value can be changed (Default: false)
      attr_accessor :mutable
      
      # If true always use the short notation
      attr_accessor :modeled
      
      # * +id+ - Weaver ID for the attribute
      # * +weaver_scope+ the Base that contains this Attribute
      # * +mutable+ if true, the value can be changed
      def initialize(id, weaver_scope, mutable = false)
        super(id, weaver_scope)
        @default = nil
        @group = nil
        @label = nil
        @mutable = mutable
        @required = true
        @type = :attribute
        @value = nil
        @modeled = false
        @implicit = true
      end
      
      # Deeply copy late bindings
      def initialize_copy(other_attribute)
        super
        
        if other_attribute.unresolved_value.kind_of?(LateBinding)
          dup = other_attribute.unresolved_value.clone
          @value = dup
        elsif other_attribute.value.kind_of?(Hash) || other_attribute.value.kind_of?(Array)
          @value = Marshal.load(Marshal.dump(Marshal.load(Marshal.dump(other_attribute.value))))
        end
      end
      
      # A Weaver attribute is said to be _implicit_ when it is set to the default value of its +type+.
      def implicit?
        return @implicit
      end
      
      # A Weaver attribute is said to be _implicit_ when it is set to the default value of its +type+.
      def implicit=(new_implicit)
        @implicit = new_implicit
      end
      
      # The value.  If the attribute models a LateBinding the value will be computed
      def value
        return @value unless @value.kind_of?(LateBinding)
        @value.weaver_eval(self)
      end

      # Set the value.
      # * +newval+ the newval, usually a String or Fixnum but it may also be a LateBinding      
      def value=(newval)
        # We don't enforce mutability constraints here, only in the DSL
        @value = newval
        @implicit = false
      end
      
      # The value.  If the attribute models a LateBinding the LateBinding will be returned for inspection
      def unresolved_value
        return @value
      end

      # * +ispass+ <code>true</code> if modeling a password      
      def password=(ispass)
        tags.add(:password) if ispass
        tags.delete(:password) if !ispass
      end
      
      # <code>true</code> if modeling a password
      def password?
        return tags.member?(:password)
      end
      
      # The language ':tags =>' modifier which adds a Symbol or list of Symbol to an attribute.
      # * *Args*    :
      #   - +newtags+ -> any Symbol or Array of Symbol; +:novalidate+ is an example of a symbol with special meaning
      def tags=(newtags)
        # Support the alternate of a single Symbol
        newtags = [newtags] if newtags.kind_of?(Symbol)
        unless newtags.respond_to?(:each)
          add_validation_failure("Syntax: tags => :<SYMBOL> | [<SYMBOL>, ...]")
          return
        end
        
        newtags.each do |newsym|
          if newsym.kind_of?(Symbol)
            tags.add(newsym)
          else
            add_validation_failure("#{newsym.inspect} is not a Symbol")
          end
        end
      end
      
      # :stopdoc:
      def to_s
#        "attribute[id: #{@id}, type: #{@type}, name: #{@name}, label: #{label}, required: #{@required}, default: #{@default}, mutable: #{@mutable}, :group: #{@group}, :value: '#{value}'"
        "attribute[id: #{@id}, type: #{@type}, name: #{@name}, label: #{label}, required: #{@required}, default: #{@default}, mutable: #{@mutable}, :group: #{@group}'"
      end
      # :startdoc:

      # Returns <code>true</code> if +another_attribute+ is equal by value.
      def ==(another_attribute)
        retval = super
        # Ignore description and other non-functional values
        return retval && @id == another_attribute.id && unresolved_value == another_attribute.unresolved_value
      end
      
      # Validates the entity and its direct and indirect contained children.
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ an optional Array which may be +[:predeployment]+
      #
      # @see Base#validate_collect
      def validate_collect(out_messages, desired_phases = nil)
        super
        
        return if tags.member?(:novalidate)
        
        # Can the attribute be evaluated?
        begin
          v = value()
          
          if (desired_phases || []).member?(:predeployment) && required && v.nil?
            out_messages.push ProblemStatus.new(Logger::WARN, "Required value #{fully_qualified_id} is nil", self)
          end
        rescue WeaverError => we
          report(Logger::WARN, "Can't evaluate late_binding #{fully_qualified_id}", self, we)
        rescue TypeError => te
          report(Logger::WARN, "Can't evaluate late_binding #{fully_qualified_id}", self, te)
        end
      end

      # Usually +[]+, but if the value is a late binding an array of the attributes whose values were used in calculations
      def attribute_dependencies
        return [] unless @value.kind_of?(LateBinding) 
        return @value.attribute_dependencies(scope).keys
      end
      
      # Returns true if the attribute is modeled in a Weaver type.  Attributes that are modeled
      # in their type can be expressed with 'x y' notation instead of 'attribute :x=>y' notation
      def modeled?
        return true if @modeled
        return false if scope.nil? || scope.provenance.nil?
        prototype_attr = scope.provenance.amd[@id]
        return !prototype_attr.nil? && prototype_attr.implicit? && prototype_attr.type == @type && prototype_attr.mutable == @mutable
      end
      
      # Set the legal values for the attribute to take
      # * +vv+ a comma-separated list of values
      def valid_values=(vv)
        raise WeaverError.new("Attempt to override valid_values") unless valid_values.nil?
        add_validator(:valid_values, *vv)
      end
      
      # What are the legal values for the attribute to take?
      # * Returns an array
      def valid_values
        vv = validators.find { |entry| entry[0] == :valid_values }
        return vv.nil? ? nil : vv.slice(1..-1)  # vv will look like [:valid_values, values...]
      end
      
      # Set the type of the value
      # * +vt+ -> a Symbol such as +:string+
      def constrain_value_type(vt)
        add_validator(:value_type, vt)
      end
      
      # +old_to_new+ Hash of Base -> Base, called after clone or Fold::fold
      def freshen_late_bindings(old_to_new)
        if unresolved_value.kind_of?(LateBinding)
          unresolved_value.freshen_declaring_scope(old_to_new)
        end
      end
      
      # Define custom JSON deserialization handler
      # +from_json_proc+ a Proc
      def from_json=(from_json_proc)
        @from_json = from_json_proc
      end

      protected
      
    end
  end
end