require 'weaver/core/base'
require 'weaver/core/node'
require 'weaver/core/automation'
require 'weaver/validation/problem_status'

module Weaver
  module Core
    # A mixin for classes such as Node, Automation and Service which may contain DependsOn
    # 
    # Author:: Michael Kalantar
    # Copyright:: IBM
    module DependsOnMixin
      include Weaver::Core

      # Let all classes that include DependsOnMixin have a method to return their dependencies
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id --> Base
          def depends_on
            # Since mixins have no initialize() to set up @depends_on, initialize @depends_on lazily
            @depends_on || [] unless frozen?
            return @depends_on || []
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_depends_on)
        baseclazz.register_validate_handler(:validate_depends_on)
        baseclazz.register_reference_freshener(:depends_on_freshener)
      end
      
      # Deeply copy @depends_on
      def initialize_copy_depends_on(other_depends_on_container)
        # At this point, self.initialize_copy should have cloned @depends_on but the values are shallow.
        # However at this point 'nodes' should be deep
        @depends_on = other_depends_on_container.depends_on.clone()
      end # def initialize_copy_depends_on
      
      # Validate @depends_on
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_depends_on(out_messages, desired_phases = nil)
        depends_on.each do |dep| 
          out_messages.push Weaver::Validation::ProblemStatus.new(Logger::WARN, "depends_on entity with invalid type: #{dep.fully_qualified_id} (#{dep.class})", self) if !valid_dependency_target_types.member?(dep.class)
        end # 
      end # def validate_depends_on
      
      # Finish the work started by initialize_copy_depends_on() and fix up the target
      def depends_on_freshener(old_to_new)
        @depends_on = depends_on.map { |d| old_to_new.member?(d) ? old_to_new[d] : d }
      end # def depends_on_freshener
      
      protected

      # Don't need a to_weaver method; this is in weaver_core_root.erb
      
      def valid_dependency_target_types
        return [ 
          Weaver::Core::Node, 
          Weaver::Core::Automation
        ]
      end # def valid_dependency_target_types      

    end # module DependsOnMixin
  end # module Core
end # module Weaver