require 'weaver/core/base.rb'

module Weaver
  module Core
    # A mixin for classes such as Application which may contain Node
    # 
    # Author:: Ed Snible
    # Copyright:: IBM
    module NodeContainerMixin
      include Weaver::Core      # Allow us to use @@logger
      
      # Let all classes that include NodeContainerMixin have a method to return their nodes
      def self.included(baseclazz)
        baseclazz.class_exec {
          # Hash of id -> Node
          def nodes
            # Since mixins have no initialize() to set up @nodes, initialize @nodes lazily
            @nodes ||= {} unless frozen?
            return @nodes || {}
          end
        }
        
        baseclazz.register_initialize_copy_handler(:initialize_copy_nodes)
        baseclazz.register_validate_handler(:validate_nodes)
        baseclazz.register_child_enumerator(:nodes)
        baseclazz.register_child_container(:node_alias)           
        baseclazz.register_reference_freshener(:alias_freshener)
        baseclazz.register_reference_enumerator(:alias_nodes)
        
      end
      
      # Deeply copy @nodes
      def initialize_copy_nodes(other_node_container)
        # At this point, self.initialize_copy should have cloned @nodes but the values are shallow.
        
        # Deeply copy the non-alias nodes.  Leave the aliases pointing to the old Node
        @nodes = other_node_container.nodes.clone
        other_node_container.unique_nodes.each do |id, node|
          dup = node.clone()
          dup.scope = self
          @nodes[id] = dup
        end
        
        # The aliases should be handled later, by alias_freshener().  (The aliases are not yet deep copies.)
      end
      
      # Validate @nodes
      # * +out_messages+ ProblemStatus will be pushed onto this for any validation message
      # * +desired_phases+ optional Array which may be +[:predefined]:
      def validate_nodes(out_messages, desired_phases = nil)
        unique_nodes.each { |id, node| node.validate_collect(out_messages, desired_phases) if node.scope == self }
      end
      
      # Finish the work started by initialize_copy_nodes() and fix up the aliases
      def alias_freshener(old_to_new)
        alias_nodes.each do |alias_id, stale_old_node|
          @nodes[alias_id] = old_to_new[@nodes[alias_id]] if old_to_new.key?(@nodes[alias_id])
        end
      end
      
      # Create an alias for newnode_candidate if it is a Node
      def node_alias(alias_id, newnode_candidate)
        return nil unless newnode_candidate.kind_of?(Node)
        nodes[alias_id] = newnode_candidate
        return newnode_candidate
      end
      
      # The actual nodes contained by this topology (not aliases for evaluating late bindings in folded topologies
      def unique_nodes
#        return nodes.select { |id, node| id == node.id && node.scope == self } 
        retval = Hash[nodes.select { |id, node| id == node.id && node.scope.equal?(self) }]  # In Ruby 1.8.x, Hash.select returns an array
        return retval 
      end
      
      # Return Hash id->node for nodes that can be used, but are not really contained by this topology 
      # (typically aliases for handling late bindings created by folding).
      def alias_nodes
        retval = nodes.reject { |id, node| id == node.id && node.scope == self }
        return retval
      end

    end
    
  end
end