require 'weaver/core/logging'

module Weaver
  # \Weaver::Deployers provides tools to make a Weaver description come into existance by deploying virtual machines and
  # configuring them to match Weaver's Bill of Materials.
  #
  # @see Plugins::IwdDeployer
  # @see Plugins::MockDeployer
  #
  # Other gems may provide additional deployers.
  module Deployers
    
    # Base class for a deployer plugin that implements the logic
    # on how to deploy an environment to a given cloud.
    #
    # * *Plugin requirements*: 
    #  - subclass Deployer (this class) 
    #
    #  - define a constant named ID (e.g., ID = :iwd_vsp) representing the 
    #    cloud type that is used in infrastructure topologies. This is the 
    #    correlation identifier between a cloud and the plugin responsible 
    #    for deploying an Environment to that particular cloud needs to declare
    #    a constant ID identifying the plugin and can be placed in one of the 
    #    following directoryies. 
    #  - place the plugin in one of the following folders:
    #    - the plugins/ folder in the this project
    #    - $HOME/.weaver/plugins/deployers
    #    - in any GEM when it is in weaver/plugins/deployers
    #
    class Deployer
      include Weaver::Core
      
      class << self
      
        def plugins
          @plugins ||= []
        end

        # This method is called by Ruby for each Deployer plugin that subclasses
        # this class. We register it in the list of plugins and can use the find
        # method below to run through the list of plugins, evalute its ID constant
        # to find the coresponding deployer
        def inherited(subclass)
          plugins << subclass
        end

        # Find a deployer with a specific id. The id is the correlation between the
        # between the deployer implementation and the cloud statement that the users 
        # specifies in the infrastructure topology. It allows user to easily implement
        # new deployers for a custom cloud they would like to deploy to.
        # 
        # * *Args*:
        #   -+id+ the id of the plugin to find (Symbol)    
        # * *Returns*:
        #   - a Deployer instance that implements a concrete deployer.      
        def find(id)
          plugins.each do |plugin|
            if plugin.respond_to? :id
              return plugin if plugin.id() == id
            else
              @@logger.warn("Deployer plugin #{plugin} does not declare a method 'self.id()' returning the topology ids it can deploy.")
            end
          end
          # we cannot find a plugin with the given id
          @@logger.warn("Cannot find a Deployer plugin with ID '#{id}'")
          return nil
        end
      end
      
      attr_reader :properties
      
      def initialize(properties = {}) 
        @properties = properties        
      end
      
      # Deploy the environment by bringing up all resources.
      #
      # * *Args*    :
      #   -+environment+ An instance of an Environment object that represents the binding
      #                              of an application to an environent.
      #   -+deploy_parameters+ External attributes that need to be set.
      # * *Returns* :
      #   -+Environment+ The updated environment instance with runtime information such as hostname and ip.
      #
      def deploy(environment, deploy_parameters = {})       
        raise WeaverError, "A deployer is expected to override this method"      
      end
    
      # Simulate the model that deployment should return.  This means, for example, creating IP addresses for
      # nodes.  We will use this to validate properties that depend on attributes propagating from cloud-provided
      # values.  This method MAY tag the returned model with additional warnings; for example if the model
      # requests a fixed IP address that will fail an actual deployment this method should tag the attribute 
      # as having a problematic value.
      #
      # * *Args*    :
      #   -+environment+ An instance of an Environment object that provides a model of an intended 
      #                     deployment and can receive ProblemStatus annotations.
      #   -+deploy_parameters+ External attributes that need to be set.
      # * *Returns* :
      #   -+Environment+ The updated environment instance with problem annotations and runtime information such as hostname and ip.
      #
      def simulate_deploy(environment, deploy_parameters = {})       
        raise WeaverError, "A Deployer is expected to override this method"      
      end
    
    end

  end # Deployers module
end # Weaver module
