require "weaver/core/logging" 
require 'weaver/chef/chef_automation_handler'  
require 'erubis'
require 'json'
require 'base64'

module Weaver
  module Deployers
    module Execution
      # Create cloud-init data. Currently, Weaver creates a part handler that handles two types of input:
      # (1) Weaver-generated bootstrap script and (2) description of continuous validation rules. 
      # Weaver expects exactly one bootstrap script, and zero or more descriptions of continuous validation rules. 
      class CloudInit
        include Weaver::Core
        
        # Returns enumeration of covered Weaver entities (in this case Automation)
        def covers(node)
          return node.flatten_children.select { |ent| ent.weaver_instance_of?(:chef_automation) }
        end
        
        # Attaches failure to node if its covered automation is invalid
        def validate(node)
          begin
            handler = Weaver::Chef::ChefAutomationHandler.new()
            # convert each chef_node_attr that is a Hash to a String
            handler.interpolate_and_convert_chef_attributes(node.scope, node)
            # If there are chef_automations, collect the attribute values to be used together in one chef automation
            # otherwise, attrs will hold default values
            attrs = handler.generate_chef_attributes(node)
          rescue WeaverError => we
            @@logger.warn "Couldn't generate Chef attributes: #{we.message}, backtrace=#{we.backtrace}"
            node.add_validation_failure(we.message)
          end
        end

        def create_userdata(node, deployment_params, urls_to_files, vm_update = false, node_to_rules = {})
          create_filtered_userdata(node, [], deployment_params, urls_to_files, vm_update, {})
        end

        # Create user_data in cloud-init format
        # * +node+ a Node
        # * +automation_filters+ list of automation ids to be selected
        # * +deployment_params+ Hash containing :bootstrap_url for bootstrap script package, :deps_file_url
        # * +urls_to_files+ Hash of url to filename, e.g. { "https://s3.amazonaws.com/weaver2012/jke-20121018-1639.zip" => "/tmp/jke-20121018-1639.zip" }
        # * +node_to_rules+ Hash of node id to the applicable continuous validation rules 
        def create_filtered_userdata(node, automation_filters, deployment_params, urls_to_files, vm_update = false, node_to_rules = {})
          bootstrap_script = create_bootstrap_script(node,automation_filters,deployment_params,urls_to_files, vm_update)
          # Fabio: The following two lines were commented out so that we no longer rely on a part handler 
          #user_data = !vm_update ? create_part_handler_and_parts(node, bootstrap_script, node_to_rules) : bootstrap_script
          #return user_data
          return bootstrap_script
        end 
        
        def validate_userdata(node, deployment_params, deps_file_url, automation_type)
          # TODO: validate that the bootstrap_url and deps_file_url can be fetched
          if automation_type == :chef_solo
            if (deps_file_url.nil? || deps_file_url.strip.empty?)
              @@logger.warn "There is no deps_file_url specified for deployment on node #{node.name}"
            end
            if (deployment_params[:bootstrap_url].nil?)
              @@logger.error "There is no bootstrap_url specified as a deploy parameter for node #{node.name}"
              raise WeaverError, "A bootstrap url is required for deployment with chef-solo yet none is specified; resolve by passing bootstrap_url as a deploy parameter"
            end
          end
        end
        
        protected
        
        def create_bootstrap_script(node, automation_filters, deployment_params, urls_to_files, vm_update = false)
          handler = Weaver::Chef::ChefAutomationHandler.new()
          # convert each chef_node_attr that is a Hash to a String
          # TODO: The following method only interpolates attribute values that are in a script_package, in a Hash. Evaluate if there is any reason to call this. Commenting out for now.
          # handler.interpolate_and_convert_chef_attributes(node.scope, node)

          # If there are chef_automations, collect the attribute values to be used together in one chef automation
          # otherwise, attrs will hold default values
          attrs = handler.generate_chef_attributes(node, automation_filters)

          # if there is a deploy parameter for deps_file_url, use that - it overrides the deps_file_url in the model. 
          # If there is no deploy parameter, use the deps_file_url specified in the model in a chef_automations
          deps_file_url = deployment_params[:deps_file_url].nil? ? attrs[:deps_file_url] : deployment_params[:deps_file_url]
          run_list = handler.get_node_runlist(node, automation_filters)

          # TODO: for now, validate for chef_solo. Factor out later.
          validate_userdata(node, deployment_params, deps_file_url, :chef_solo)

          user = ""
          if deployment_params[:bootstrap_userid] && deployment_params[:bootstrap_password]
            user = "--user #{deployment_params[:bootstrap_userid]}:#{deployment_params[:bootstrap_password]} "
          end

          eruby = Erubis::Eruby.new(File.read(File.join(File.dirname(__FILE__), '..', '..', '..', '..', 'data/templates/bootstrap_script.erb')))
          bootstrap_script = eruby.result({:deployment_params=>deployment_params, :urls_to_files=>urls_to_files, 
            :chef_node=>node.id.to_s, :chef_node_attr=>attrs[:chef_node_attr], :deps_file_url=>deps_file_url, :run_list=>run_list, :user=>user, :vm_update=>vm_update})
          return bootstrap_script   
        end
        
        ####
        #### Fabio: The methods below are no longer used. They used to create a multi-part MIME cloud-init user-data and our
        #### Weaver part handler. We no longer use this approach to generating cloud-init user data.
        #### I am keeping them here for a while... 
        ####
        
#        def create_part_handler_and_parts(node, bootstrap_script, node_to_rules = {})
#          eruby = Erubis::Eruby.new(File.read(File.join(File.dirname(__FILE__), '..', '..', '..', '..', 'data/templates/cloud_init_userdata.erb')))
#          user_data = eruby.result({:encoded_bootstrap_script => Base64.encode64(bootstrap_script), :artifacts => create_artifact_parts_payloads(node, node_to_rules), :rules => create_rule_parts_payloads(node,node_to_rules)})
#          return user_data
#        end
#           
#        def create_artifact_parts_payloads(node, node_to_rules)
#          # Determine which artifacts need to be included in this node, based on the artifacts referenced by the applicable rules
#          selected_rules = node_to_rules[node.id]
#          referenced_artifacts = []
#          selected_rules.each do |r|
#            referenced_artifacts.concat(r.contval.artifacts)
#          end
#          @@logger.debug "Artifact ids referenced by rules applicable to node :#{node.id} --> #{referenced_artifacts.inspect}"
#          needed_artifacts = node.root_entity.artifacts.values_at(*referenced_artifacts)
#          @@logger.debug "Artifacts for which user-data parts will be created: #{needed_artifacts.inspect}"
#          
#          # Compute a Hash of artifact id to the corresponding JSON part payload
#          # We will have one part per artifact needed by this node
#          retval = {}
#          needed_artifacts.each do |artifact|
#            artifact_hash = {}
#            artifact_hash["embedded"] = (artifact.location == :local ? true : false)
#            artifact_hash["target"] = artifact.target_path
#            artifact_hash["exec"] = artifact.exec
#            if artifact.location == :local
#              artifact_local_path = File.join(Dir.pwd, 'rules', 'artifacts', artifact.source_path)
#              @@logger.debug "Artifact :#{artifact.id} local path: #{artifact_local_path}"
#              file = File.open(artifact_local_path, "r")
#              contents = file.read()
#              file.close()
#              artifact_hash["artifact"] = Base64.encode64(contents)
#            end
#            # TODO Handle artifacts stored in the Weaver service
#            #else
#            part_payload = JSON.pretty_generate(artifact_hash)
#            @@logger.debug "Part payload for artifact :#{artifact.id} --> #{part_payload}"
#            retval[artifact.id] = part_payload
#          end
#          return retval
#        end
#        
#        def create_rule_parts_payloads(node, node_to_rules)
#          # Compute a Hash of rule id to the corresponding JSON part payload
#          # We will have one part per rule needed by this node
#          retval = {}
#          node_to_rules[node.id].each do |rule|
#            rule_hash = {}
#            rule_hash["type"] = rule.contval.runtime_type
#            rule_hash["period"] = rule.contval.period
#            rule_hash["command"] = rule.contval.command
#            part_payload = JSON.pretty_generate(rule_hash)
#            @@logger.debug "Part payload for rule :#{rule.id} --> #{part_payload}"
#            retval[rule.id] = part_payload
#          end
#          return retval
#        end
      end
    end
  end
end

