require 'weaver/core/base.rb'
require 'timeout'

module Weaver
  module Providers
    # A provider that merely deploys inner Topology s.  It cannot deploy Node s.
    # @see Weaver::Deployers::Deployer
    #
    # Author:: Ed Snible
    # Copyright:: IBM
    class DelegatingDeployer < Weaver::Deployers::Deployer
      
      def initialize(properties = {}, &reporter)
        super
        
        @reporter = reporter || Proc.new { |lvl, status, msg| }
      end
      
      def deploy(topology, deploy_parameters = {})
        # 'Predeploy' topology-level services
        # The services should have at least access to :weaver_coordination_topology_id and :mock
        
        topology.services.each_value { |service|
          service.predeploy(deploy_parameters)
        }
        
        # TODO: this code takes only care of services on the topology - Ed is this correct?
        begin
          Timeout::timeout(300) {
            topology.services.each_value { |service| service.join }
          }
        rescue Timeout::Error => e
          puts caller
          raise WeaverError, "Execution of service methods timed out"
        end
          
        # Deploy topology-level services
        # The services should have at least access to :weaver_coordination_topology_id and :mock
        topology.services.each_value { |service|
          service.deploy(deploy_parameters)
        }
          
        # We want to recursively call ourself on any providerless topologies
        topology.topologies.each do |childid, child| 
          if child.provisioner.nil?
            deployed_top = deploy(child, deploy_parameters)
          else
            provisioner = child.provisioner
            deployer = provisioner.deployer(@properties, &@reporter)
            deployer.deploy(child, deploy_parameters)
          end
        end
          
        # Warn about any non-alias Nodes in topology
        topology.nodes.values.select { |node| node.scope.equal?(topology) }.each do |node|
          prov = node.provisioner
          if prov.nil?
            raise WeaverError.new("Aborted") if @reporter.call(Logger::WARN, "Deploying", "#{node.fully_qualified_id} has no provider") == :abort
          else
            raise WeaverError.new("Aborted") if @reporter.call(Logger::WARN, "Deploying", "Warning: #{node.fully_qualified_id} cannot be deployed because it in a topology with no provider") == :abort
          end
        end
        
        begin
          Timeout::timeout(300) {
            topology.services.each_value { |service| service.join }
          }
        rescue Timeout::Error => e
          puts caller
          raise WeaverError, "Execution of service methods timed out"
        end
        
        # Deploy topology-level services
        # The services should have at least access to :weaver_coordination_topology_id and :mock
        topology.services.each_value { |service|
          service.postdeploy(deploy_parameters)
        }
        begin
          Timeout::timeout(300) {
            topology.services.each_value { |service| service.join }
          }
        rescue Timeout::Error => e
          puts caller
          raise WeaverError, "Execution of service methods timed out"
        end
      end
      
      def simulate_deploy(topology, deploy_parameters = {})
               
        # We want to recursively call ourself on any providerless topologies
        topology.topologies.each do |childid, child| 
          if child.provisioner.nil?
            simulate_deploy(child, deploy_parameters)
          else
            provisioner = child.provisioner
            deployer = provisioner.deployer(@properties, &@reporter)
            deployer.simulate_deploy(child, deploy_parameters)
          end
        end
        
        # Warn about any providerless nodes in providerless topologies
        if topology.provisioner.nil?
          topology.nodes.values.select { |node| node.scope.equal?(topology) && node.ip_address.nil? && node.hostname.nil? }.each do |node|
            prov = node.provisioner
            if prov.nil?
              @@logger.debug "#{node.fully_qualified_id} has no provider"
              node.report(Logger::WARN, "Node has no provider", node)
            else
              @@logger.debug "Node cannot be deployed because it in a topology with no provider"
              node.report(Logger::WARN, "#{node.fully_qualified_id} cannot be deployed because it in a topology with no provider", node)
            end
          end
        end
        
      end
      
    end
  end
end