require 'weaver/core/logging'
require 'weaver/transforms/duplicate'
require 'set'

module Weaver
  module Transforms

    # Implements Weaver 'topology folding'.
    #
    # Author:: Ed Snible
    class Fold
      include Weaver::Core
      
      # @private
      # We avoid folding name and description because they are non-functional and
      # we wish to preserve the original names.  We avoid folding :ip_address and :hostname
      # because they aren't 'true' attributes but rather aliases for :nic0's attributes.
      UNFOLDED_ATTRIBUTE_IDS = Set.new [:name, :description, :ip_address, :hostname ]

      # Create an entity logically equivelent but containing no realized_by relations
      # * +entity+ a {Base}, typically a {Topology} or {Root}, to be folded    
      def self.fold(entity)
        raise WeaverError.new("Can't fold nil") if entity.nil?
        
        folder = Fold.new(entity)
        return folder.folded_entity
      end
      
      # The actual folded value
      def folded_entity
        return @entity
      end
      
      protected
      
      # * +orig_entity+ a Base, typically a Topology or Root, to be folded    
      def initialize(orig_entity)
        @entity = Duplicate.duplicate(orig_entity)
        
        # If we are replacing anything, record it here so references can be repointed to the new thing
        @old_to_new = {}
        @old_to_all_new = Hash.new { |h, k| h[k] = [] }
          
        # After realization what values will mutable attributes be expected to hold?
        @final_attribute_values = {}
          
        @entity.flatten_children.
          select { |entity| entity.kind_of?(Node) }.each do |node|
            node.realized_by.each { |target| ensure_valid_realization(node, target) } 
          end
            
#        # TODO Currently we don't ensure that merging topology attributes won't lead to value clashes
        
        @entity.flatten_children.
          select { |entity| entity.kind_of?(Node) }.each do |node| 
            node.realized_by.each { |target| merge_nodes(node, target) }
          end
          
        retarget_entity(@entity)
        @entity.flatten_children.each { |child| retarget_entity(child) }
        
        @entity.update_references(@old_to_new)
      end
      
      # RECURSIVE
      def build_mapping(new_entity, orig_entity)
#        if @old_to_new.member?(orig_entity)
#          raise WeaverError.new("Original #{orig_entity.fully_qualified_id} maps to both #{@old_to_new[orig_entity].fully_qualified_id} and #{new_entity.fully_qualified_id}")
#        end
        
        @old_to_new[orig_entity] = new_entity unless @old_to_new.member?(orig_entity)
        @old_to_all_new[orig_entity] << new_entity
        
        orig_children = orig_entity.contained_entities
        new_children = new_entity.contained_entities 
        new_children.each do |id, new_child|
          raise WeaverError.new("Missing child #{id.inspect} on #{orig_entity.id.inspect}") if orig_children[id].nil? 
          build_mapping(new_child, orig_children[id]) 
        end
      end
      
      # A realization is said to be 'valid' if source's attributes are a subset of target's and
      # if any Components on Source that are not to-be-installed-with-a-Script have a match
      # on target.
      # * +source+ a Node
      # * +target+ a Node
      def ensure_valid_realization(source, target)
        unless target.kind_of?(Node)
          raise WeaverError.new("#{source.fully_qualified_id} is not realized by a known node")
        end
        
        ensure_nic_realization(source, target)
  
        source.amd.each do |id, attr|
          ensure_attribute_realization(attr, 
            target.amd[id])
        end
        
        source.components.each do |id, component|
          ensure_component_realization(component, target.components[id])
        end
        
        # TODO in the future we could insist on only valid policy realizations here, or generalize ensure_valid_realization
        # by moving it into Base to help Fold::fold be made aware of model elements.
      end
  
      # Can the source Component be realized upon the target without an attribute value conflict?      
      # * +source_comp+ a Component that is realized by +target_comp+
      # * +target_comp+ a Component that realizes +source_comp+
      def ensure_component_realization(source_comp, target_comp)
        return if target_comp.nil?
        
        source_comp.amd.each do |id, attr|
          ensure_attribute_realization(attr, target_comp.amd[id])
        end
      end

      # Can the source Node be realized upon the target without an IP Address conflict?      
      # * +source+ a Node that is realized by +target+
      # * +target+ a Node that realizes +source+
      def ensure_nic_realization(source, target)
        # This implementation only checks :nic0
        source.nics[:nic0].amd.each do |id, attr|
          ensure_attribute_realization(attr, target.nics[:nic0].amd[id])
        end
      end
      
      # Can the source Attribute be realized upon the target without a value conflict?      
      # * +source_attr+ an Attribute whose container is realized by +target_attr+'s container
      # * +target_attr+ an Attribute whose container realizes +source_attr+'s container
      def ensure_attribute_realization(source_attr, target_attr)
        return if source_attr.implicit? || UNFOLDED_ATTRIBUTE_IDS.member?(source_attr.id)
    
        if target_attr.nil?
          raise WeaverError.new("Attribute :#{source_attr.id} (='#{source_attr.value}') doesn't exist on target")
        end
    
        return if target_attr.value == source_attr.value
  
        unless target_attr.mutable
          raise WeaverError.new("Attribute :#{source_attr.id} mismatch; source is '#{source_attr.value}', target is '#{target_attr.value}' and #{target_attr.fully_qualified_id} is not mutable")
        end
      
        # It isn't sufficient that the target attribute be mutable if multiple Nodes will overwrite
        # the target value with *different* values
        if @final_attribute_values.key?(target_attr)
          other_value = @final_attribute_values[target_attr]
          raise WeaverError.new("Attribute :#{source_attr.id} clashes; folding requires it to be both #{other_value} and #{target_attr.value}") unless other_value == source_attr.value
        else
          @final_attribute_values[target_attr] = source_attr.value
        end
      end

      # Make +target+ look like +source+ so that the realization will be valid
      # * +source+ a Node that is realized by +target+
      # * +target+ a Node that realizes +source+
      def merge_nodes(source, target)
        raise WeaverError.new("realized_by cycle: #{source.fully_qualified_id }") if source.equal?(target) 
      
        merge_attributes(source, target)
        merge_validators(source, target)
        merge_requirements(source, target)
        merge_nics(source, target)
        merge_tags(source, target)
        
        # Merge components that exist on source but not on target
        source.components.each do |id, source_component|
          target_component = target.components[id]
          if target_component.nil?
            dup = source_component.clone()
            # @@logger.debug "Fold.merge_nodes creating duplicate Component #{source_component.id.inspect}"
            dup.scope = target
            build_mapping(dup, source_component)
            target.components[id] = dup
          else
            merge_attributes(source_component, target_component)
            merge_validators(source_component, target_component)
          end
        end
        
        # Merge automation that exists on source but not on target        
        source.automation.each do |id, source_automation|
          target_automation = target.automation[id]
          if target_automation.nil?
            dup = source_automation.clone()
            dup.scope = target
            build_mapping(dup, source_automation)
            target.automation[id] = dup
          else
            merge_attributes(source_automation, target_automation) 
            merge_validators(source_automation, target_automation)
          end
        end
        
        # Merge policies that exist on source but not on target
        source.policies.each do |id, source_policy|
          target_policy = target.policies[id]
          if target_policy.nil?
            dup = source_policy.clone
            dup.scope = target
            build_mapping(dup, source_policy)
            target.policies[id] = dup
          else
            merge_attributes(source_policy, target_policy) 
            merge_validators(source_policy, target_policy)
          end
        end
        
        # Merge services that exist on source but not on target
        source.services.each do |id, source_service|
          target_service = target.services[id]
          if target_service.nil?
            dup = source_service.clone()
            # @@logger.debug "Fold.merge_nodes creating duplicate Service #{source_service.id.inspect}"
            dup.scope = target
            build_mapping(dup, source_service)
            target.services[id] = dup
          else
            merge_attributes(source_service, target_service)
            merge_validators(source_service, target_service)
          end
        end
        
        # Merge automation_order arrays (assumption is that the target runs first, then the source)
        target.automation_order.push(source.automation_order).flatten!
        
        merge_depends_on(source,target)     
      end
     
      # Merge validators
      # * +source+ a Node that is realized by +target+ | one of the node's Components
      # * +target+ a Node that realizes +source+ | one of the node's Components
      def merge_validators(source, target)
        # (Currently we ignore the unlikely possibility of removing duplicates)
        target.validators.concat(source.validators)
      end
      
      # * +source+ a Node that is realized by +target+
      # * +target+ a Node that realizes +source+
      def merge_requirements(source, target)
        source.requirements.each { |requirement| target.requirements << requirement }
      end
      
      # * +source+ a Node that is realized by +target+
      # * +target+ a Node that realizes +source+
      def merge_nics(source, target)
        unless source.nics[:nic0].ip_address.nil?
          target.nics[:nic0].ip_address = source.nics[:nic0].ip_address
          target.nics[:nic0].amd[:ip_address].weaver_origin = source.nics[:nic0].amd[:ip_address].weaver_origin
        end
      end
      
      # Merge attributes if mutable
      # * +source+ a Node or Component
      # * +target+ a Node or Component
      def merge_attributes(source, target)
        
        # Don't bother merging attribute meta data, but merge values and origin
        source.amd.each do |id, attr|
          # Don't fold :name and description, or nil-valued attributes
          unless UNFOLDED_ATTRIBUTE_IDS.member?(attr.id) || attr.implicit?
#            target.make_attribute_explicit(id) if target.amd[id].implicit?
            target_attr = target.amd[id]
            raise WeaverError.new("Source #{source.type} #{source.id} has attribute #{id} but Target #{target.type} #{target.id} does not") if target_attr.nil?  
            if target_attr.mutable
              target_attr.value = attr.value
              target_attr.weaver_origin = attr.weaver_origin
            end
          end
        end
      end
      
      # Merge tags
      # * +source+ a Node or Component
      # * +target+ a Node or Component
      def merge_tags(source,target)
        source.tags.each { |tag| target.tags << tag } 
      end # def merge_tags
      
      # Merge dependencies
      # * +source+ a Node or Component
      # * +target+ a Node or Component
      def merge_depends_on(source,target)
        source.depends_on.each { |d| target.depends_on << d unless target.depends_on.member?(d) }
      end # def merge_depends_on
      
      # Find any bound Node that is a realization of the given Node
      # RECURSIVE
      # * +node+ a Node
      # * +accum+ Set of Nodes
      def a_final_realization(node, accum = nil)
        return node if node.nil? || node.realized_by.empty?
        
        accum = Set.new if accum.nil?
        raise WeaverError.new("realized_by cycle: #{accum.to_a.map { |entity| entity.fully_qualified_id }.inspect}") if accum.member?(node) 
        accum.add(node) 
        return a_final_realization(node.realized_by[0], accum)
      end
      
      # RECURSIVE
      # * +node+ a Node
      # * +result_accum+ Set of bound Nodes that directly or indirectly realize a Node
      # * +intermediate_accum+ Set of unbound Nodes that directly or indirectly realize a Node; the set includes the Node itself
      def all_final_realizations_accum(node, result_accum, intermediate_accum)
        raise WeaverError.new("Unknown node #{node} (a #{node.class})") unless node.kind_of?(Node)  # It might be a Symbol
        
        # Is +node+ bound?
        if node.realized_by.empty?
          result_accum.add(node)
        else
          raise WeaverError.new("realized_by cycle: #{result_accum.to_a.map { |entity| entity.fully_qualified_id }.inspect}") if intermediate_accum.member?(node)
          intermediate_accum.add(node)
          
          node.realized_by.each { |realization| all_final_realizations_accum(realization, result_accum, intermediate_accum) }
        end
      end
      
      # Find all bound Nodes that are realizations of the given Node
      # * +node+ a Node
      def all_final_realizations(node)
        accum = Set.new
        all_final_realizations_accum(node, accum, Set.new)
        return accum.to_a
      end
      
      # Given an Array of Node|Array-of-Node, find a mapping to the equivalent Array of bound Nodes
      # RECURSIVE
      # * +nodes+ array of Node | Array of Node
      def final_realizations(nodes)
        retval = []
        nodes.each do |ele|
          # If this is an Array of nodes starting together, call recursively
          if ele.respond_to?(:map)
            retval << final_realizations(ele)
          else
            bound_replacements = all_final_realizations(ele)
            if bound_replacements.length == 1
              # No fan-out, use the realization (which, for bound nodes, is the identity)
              retval << bound_replacements[0]
            else
              # Fan-out for +ele+, in the folded topology represent it by a parallel startup of the realized nodes
              retval << bound_replacements
            end
          end
        end
        return retval
      end

      # Given a container such as a Topology containing unbound Nodes, retarget the +nodes+ Hash so that
      # it is aliases to bound nodes.  Do the same for +startup_order+.
      # Furthermore, to make it easier to write +late_binding+s, do something
      # similar for Users and Attributes.
      # * +entity+ a Base
      def retarget_entity(entity)
        if entity.respond_to?(:nodes)
          newcontainers = Set.new
          newnodes = {}
          entity.nodes.each do |id, oldnode|
            
            node = a_final_realization(oldnode)
            if !node.equal?(oldnode)
              
              # Create an alias on the target
              if node.scope.nodes[id].nil?
                node.scope.nodes[id] = node
                newcontainers.add(node.scope)
              end
              
              # @@logger.debug "Fold.retarget_entity replacing Node #{oldnode.id.inspect} with its final realization #{node.id} and remembering #{oldnode.object_id} ---> #{node.object_id}"
              @old_to_new[oldnode] = node
              @old_to_all_new[oldnode] << node
            end
            
            newnodes[id] = node
         end
        
          entity.nodes.clear
          entity.nodes.merge!(newnodes)
          
          # Instead of an explicit realized_by for users, all Users get copied upon a Topology if any sibling Nodes were copied
          if entity.respond_to?(:users)
            entity.users.each do |id, olduser|
              newcontainers.each do |target_topology|
                dup = olduser.clone
                dup.scope = target_topology
                target_topology.users[id] = dup
              end
            end
          end
          
          # Instead of an explicit realized_by for attributes, all Attributes get copied upon a Topology if any sibling Nodes were copied
          newcontainers.each { |target_topology| merge_attributes(entity, target_topology) }
        end
        
        if entity.respond_to?(:startup_order)
          entity.startup_order = final_realizations(entity.startup_order)
        end
      end
      
    end      
  end
end
