require 'weaver/core/logging'

module Weaver
  module Utils

    # This plugin loader implementation combines and unifies two previous versions of
    # plugin loaders, used for the Weaver CLI and for Cloud deployers.
    #
    # Author:: Waldemar Hummer
    #
    # This class loads Weaver plugins based on a lookup.
    # Objects of this class are initialized with a list of relative file path, say, REL_PATHS,
    # and with a list of absolute paths, say, ABS_PATHS.
    # The plugin lookup order is as follows (later entries overwrite earlier entries):
    # - in any GEM when it is in any path in $REL_PATHS
    # - in any entry on the LOAD_PATH when it is in any path in $REL_PATHS
    # - in any path in $ABS_PATHS
    # - in the home directory path $HOME/.$REL_PATH for each path $REL_PATH in $REL_PATHS (e.g., /home/user/.weaver/plugins/cli)
    class PluginLoader
      include Weaver::Core

      # Instantiates a plugin loader
      # * *Args*    :
      # - +relative_search_paths+ -> an array of strings containing relative file paths
      #   that will be used to look up plugins. For instance, if this parameter
      #   is passed as ["weaver/plugins/cli/"], then we will look for plugins in 
      #   $LPE/weaver/plugins/cli/ (where $LPE is any LOAD_PATH entry), and
      #   $HOME/.weaver/plugins/cli
      def initialize(relative_search_paths, absolute_search_paths = nil)
        @relative_search_paths = relative_search_paths
        if @relative_search_paths.kind_of?(String)
          @relative_search_paths = [ @relative_search_paths ]
        end
        @absolute_search_paths = absolute_search_paths.nil? ? [] : absolute_search_paths
        if @absolute_search_paths.kind_of?(String)
          @absolute_search_paths = [ @absolute_search_paths ]
        end
      end

      # Load all the commands
      def load_plugins
        subcommand_files.each { |subcommand|
          begin
            Kernel.require subcommand
          rescue Exception => ex
            puts "Failed load of #{subcommand} because of #{ex} at #{ex.backtrace}"
            @@logger.warn "Failed load of #{subcommand} because of #{ex} at #{ex.backtrace}"
          end
        }
        return true
      end

      def subcommand_files
        @subcommand_files = Set.new
        @subcommand_files.merge gem_subcommands
        @subcommand_files.merge find_subcommands_via_loadpath
        @subcommand_files.merge find_subcommands_via_dirglob
        @subcommand_files.merge find_subcommands_via_homedir
        #puts @subcommand_files.inspect
        @subcommand_files
      end

      # Returns a Set of paths to Weaver plugins located in the user's home directory
      def find_subcommands_via_homedir
        user_specific_files = Set.new
        if ENV['HOME']
          @relative_search_paths.each do |search_path|
            Dir.glob(File.join(ENV['HOME'], "." + search_path, '*.rb')).each do |new_file|
              user_specific_files << File.expand_path(new_file)
            end
          end
        end
        user_specific_files
      end

      def gem_subcommands
        # search all gems for the given search paths
        require 'rubygems'
        find_subcommands_via_rubygems
      rescue LoadError
        Set.new
      end

      def find_subcommands_via_dirglob
        subcommand_files = Set.new
        @absolute_search_paths.each do |search_path|
          Dir.glob(File.join(search_path, '*.rb')).each do |new_file|
            #puts "new file find_subcommands_via_dirglob #{new_file}"
            subcommand_files << File.expand_path(new_file)
          end
        end
        subcommand_files
      end

      # TODO: needed?
      def find_subcommands_via_rubygems
        subcommand_files = Set.new
        @relative_search_paths.each do |search_path|
          files = Gem.find_files File.join(search_path, "*.rb")
          files.each do |file|
            #puts "gem file #{file}"
            subcommand_files << File.expand_path(file)
          end
        end
        subcommand_files
      end

      def find_subcommands_via_loadpath(paths = [File.join("weaver", "cli", "plugins")])
        files = Set.new
        $LOAD_PATH.each do |p|
          #puts "--> #{p}"
          @relative_search_paths.each do |search_path|
            path = File.join(p, search_path)
            Dir[File.expand_path("#{path}#{File::SEPARATOR}*.rb")].each do |file|
              files << File.expand_path(file.to_s)
            end
          end
        end
        #puts "loadpath files: #{files.inspect}"
        files
      end
    end
  end
end