require 'weaver/core/status_constants'

module Weaver
  # \Weaver::Validation provides data structures for validating Weaver models.
  #
  # To validate a \Weaver model, call {Weaver::Core::Abstract#validate_collect}.
  module Validation
    # Represents something the validator can generate
    class ProblemStatus
      include Weaver::Core
      
      # a Logger level; Logger::ERROR is usual
      attr_reader :level
      
      # an untranslated message
      attr_reader :message
      
      # a Weaver entity associated with the message
      attr_reader :context
      
      # a file:lineno if different than the line declaring 'context'
      attr_accessor :origin
      
      # optional special type (e.g. an IWD problem)
      attr_accessor :type
      
      # optional additional details
      attr_accessor :description
      
      # optional array of Resolutions with evaluated strings
      attr_accessor :resolutions
      
      # optional exception
      attr_accessor :exception
      
      # Represents a model problem
      # * *Args*    :
      #   - +level+ -> a Logger level; Logger::ERROR is usual
      #   - +message+ -> an untranslated message
      #   - +context+ -> a Weaver entity associated with the message
      #   - +origin+ -> a file:lineno if different than the line declaring 'context'
      #   - +type+ -> optional special type (e.g. an IWD problem)
      #   - +exception+ -> optional exception
      #   - +description+ -> optional additional details
      #   - +resolutions+ -> optional array of Resolutions with evaluated strings
      def initialize(level, message, context, origin=nil, type=nil, exception=nil, description=nil, resolutions=[])
        raise WeaverError.new("'context' must be a Base but it is a #{context.class}, caller=#{caller}") unless context.kind_of?(Base)
        raise WeaverError.new("'message' must be a String but it was #{message}, a #{message.class}") unless message.kind_of?(String)
        raise WeaverError.new("'type' must be a String but it is #{type}, a #{type.class}") unless type.nil? || type.kind_of?(String)
        raise WeaverError.new("'exception' must be an Exception") unless exception.nil? || exception.kind_of?(Exception)
        raise WeaverError.new("'description' must be a String but it is #{description}, a #{description.class}") unless description.nil? || description.kind_of?(String)
        raise WeaverError.new("'resolutions' must be an Array but it is #{resolutions}, a #{resolutions.class}") unless resolutions.nil? || resolutions.kind_of?(Array)
        
        @level = level
        @message = message
        @context = context
        @origin = !origin.nil? ? origin : first_origin(@context)
        @type = type
        @exception = exception
        @description = description
        @resolutions = resolutions || []
        
        level_to_string(level)  # Verify level is valid
      end
      
      def inspect       
        return to_s
      end      
      
      def add_resolution(label, description=nil, type=Weaver::Core::ResolutionTypes::DEFAULT, fix_proc=nil)
        @resolutions << Resolution.new(label, description, type, fix_proc)
      end
      
      # Summarize for validation report
      def to_s
        type_str=""
        type_str=" #{@type}" unless @type.nil?
        msg = "#{msg}#{@origin}: " unless @origin.nil?
        msg = "#{msg}#{level_to_string(@level)}#{type_str} in #{@context.type} :#{@context.fully_qualified_id} #{@message}"
        msg = "#{msg} at #{@origin}" unless @origin.nil?
        msg = "#{msg}; Exception: #{@exception.message}" unless @exception.nil?
        msg = "#{msg} (#{@description})" unless @description.nil?
        if !resolutions.nil? && resolutions.size > 0   
          msg = "#{msg}\n\tSuggested resolution"
          if resolutions.size > 1
            msg << "s"
          end
          msg << ":"
          resolutions.each { | resolution |
            msg = "#{msg}\n\t\t#{resolution}" unless resolution.nil?
          }
        end
        msg = "#{msg}\n"
          
        return msg
      end

      protected
            
      private
      def level_to_string(level)
        case level
        when Logger::FATAL
          return "FATAL"
        when Logger::ERROR
          return "ERROR"
        when Logger::WARN
          return "WARNING"
        when Logger::INFO
          return "INFO"
        when Logger::DEBUG
          return "DEBUG"
        else
          raise WeaverError.new("Unexpected level #{level}, use Logger::ERROR")
        end
      end
      
      # Look for weaver_origin walking up entity toward the root
      #
      # RECURSIVE
      def first_origin(entity)
        return nil if entity.nil?
        return entity.weaver_origin unless entity.weaver_origin.nil?
        return first_origin(entity.scope)
      end
      
    end
  end
end
