################################################################################
# Licensed Materials - Property of IBM Copyright IBM Corporation 2012. All Rights Reserved.
# U.S. Government Users Restricted Rights - Use, duplication or disclosure restricted by GSA ADP
# Schedule Contract with IBM Corp.
################################################################################

require 'aquarium'
require 'json'
require 'weaver_coordination'
require 'set'
require 'timeout'

module Weaver
  module Coordination
        
    # Performs AOP-based instrumentation of the Chef runtime and
    # for determining when Chef attributes are read and written to
    # achieve proper coordination for multi machine software stack
    # deployment.
    #
    # Author: Florian Rosenberg
    #
    class ChefInterceptor
      include Weaver::Coordination::Logging
      include ::Aquarium::Aspects 
      
      IGNORED_WRITES = ['hostname', 'ipaddress', 'ipaddressv6']

      # Args:
      #  +node_id+ -> the name of the node that is a child of the topology node identified by +topology_name+
        # MK - move dependency configuration to coordinator - START
    # # def initialize(coordinator_hosts = [ "localhost" ], coordinator_port = "2181", coordination_id = "", node_id = "", fq_node_id = "", read_dependencies = [], write_dependencies = [], role_dependencies = {}, timeout = 3600)
      def initialize(coordinator_hosts = [ "localhost" ], coordinator_port = "2181", coordination_id = "", node_id = "", fq_node_id = "", timeout = 3600)
        # MK - move dependency configuration to coordinator - END
        @coordinator_hosts = coordinator_hosts
        @coordinator_port = coordinator_port
        @coordination_id = coordination_id
        @node_id = node_id
        @fq_node_id = fq_node_id
        # MK - move dependency configuration to coordinator - START
#        # Convert next several from ImmutableMash to Hash to avoid intercepting our own data
#        @read_dependencies = read_dependencies.map { |d| convert_dependency(d) }
#        @write_dependencies = write_dependencies.map { |d| convert_dependency(d) }
#        @role_dependencies = role_dependencies.inject({}) { |hash,(k,v)| hash[k] = v; hash}
          # MK - move dependency configuration to coordinator - END
        @timeout= timeout
          
        @@logger.info "Using coordinator host: #{@coordinator_hosts.inspect}"
        @@logger.info "Using coordinator port: #{@coordinator_port}"
        @@logger.info "Coordination transaction id: #{@coordination_id}"
        @@logger.info "Node id: #{@node_id}"
        @@logger.info "Fully qualified node id: #{@fq_node_id}"
          # MK - move dependency configuration to coordinator - START
#        @@logger.info "Parameter dependencies (read):"
#        @read_dependencies.each {|dep| @@logger.info "   #{log_dependency(dep)}" } 
#        @@logger.info "Parameter dependencies (write):"
#        @write_dependencies.each {|dep| @@logger.info "   #{log_dependency(dep)}" }
#        @role_dependencies.each { |role,dep| @@logger.info "#{role} depends on #{dep.join(', ')}" }
          # MK - move dependency configuration to coordinator - END
        @@logger.info "Timeout: #{timeout}"

        Weaver::Coordination::System.instance.set_property(:zookeeper_hosts, @coordinator_hosts)
        Weaver::Coordination::System.instance.set_property(:zookeeper_port, @coordinator_port)
        @coordinator_instance = CoordinatedTransactionManager.instance.join_transaction(@coordination_id)
        
        # keep track of the arguments; ie, 'x' and 'y' in node['x']['y] for read and write
        # references.  Keep them separate so that we don't get mixed up accidently.
        @arguments_write = []
        @arguments_read = []
        
        # list of writes to execute.  Queued so get last value written 
        @write_queue = {}
        @run_list_trace = {}
        @current_recipe = nil
        @current_role = nil
        # MK - work in progress to move dependency details to coordinator START
        @read_dependencies = Kernel.eval(@coordinator_instance.read_data("#{@node_id}.read_dependencies"))
        @write_dependencies = Kernel.eval(@coordinator_instance.read_data("#{@node_id}.write_dependencies"))
        @role_dependencies = Kernel.eval(@coordinator_instance.read_data("#{@node_id}.role_dependencies"))
        @@logger.info "Parameter dependencies (read):"
        @read_dependencies.each {|dep| @@logger.info "   #{log_dependency(dep)}" } 
        @@logger.info "Parameter dependencies (write):"
        @write_dependencies.each {|dep| @@logger.info "   #{log_dependency(dep)}" }
        @role_dependencies.each { |role,dep| @@logger.info "#{role} depends on #{dep.join(', ')}" }
        # MK - work in progress to move dependency details to coordinator END
      end

      # Updates the node basic properties in the topology managed by Zookeeper. This is mostly
      # network related stuff (NICs and their IPs, hostnames). It builds up the following structure
      # in the coordinator:
      # /<topology_name>
      # ---/<node_name>
      # ------/ipaddress (copy of default nic information below)
      # ------/hostname  (or default IP if hostname is 'localhost')
      # ------/nics
      # ---------/en0
      # -----------/ipaddress
      # -----------/ipaddress_v6
      # ---------/en1
      # -----------/ipaddress
      # -----------/ipaddress_v6
      #
      # * *Args*    :
      #   +chef_node+  the name of the topology node in coordinator
      def update_node_properties(chef_node)
        
        # set the VM_STATUS to running -- TODO should be done by the cloud
        @coordinator_instance.write_data("#{@fq_node_id}.#{VM_STATUS_STR}", VM_STATUS::RUNNING)  
        
        # set status of the chef run to configuring
        @coordinator_instance.write_data("#{@fq_node_id}.#{Weaver::Coordination::CONFIG_STATUS_STR}", 
          Weaver::Coordination::CONFIG_STATUS::CONFIGURING)
         
        # set hostname 
        hostname = chef_node['fqdn'] 
        if hostname =~ /localhost/ or hostname.nil? or hostname.empty?
          hostname = chef_node['ipaddress']
        end
        @@logger.info "Hostname: #{hostname}"      
        @coordinator_instance.write_data("#{@fq_node_id}.hostname", hostname)

        # set nics and their IPv4 and IPv6 addresses  
        networks = chef_node.automatic['network']
        default_iface = networks['default_interface']
        default_iface_ipv6 = networks['default_inet6_interface']
        @@logger.info "Default interface: #{default_iface}"      
        @@logger.info "Default interface v6: #{default_iface_ipv6}"                    
        networks['interfaces'].each do |iface|
          iface_id = iface[0]
          encapsulation = iface[1]['encapsulation']
          unless encapsulation != 'Ethernet'
            @@logger.info "Processing NIC '#{iface_id}'"              
            # read out inet6 and inet address
            iface[1]['addresses'].select { |k,v| v['family'] == 'inet' }.each do |k, v|
              @@logger.info "  IPv4: #{k}"              
              @coordinator_instance.write_data("#{@fq_node_id}.nics.#{iface_id}.ipaddress", k)
              if iface_id == default_iface
                @coordinator_instance.write_data("#{@fq_node_id}.ipaddress", k)
                @coordinator_instance.write_data("#{@fq_node_id}.ip_address", k)
              end
            end
            iface[1]['addresses'].select { |k,v| v['family'] == 'inet6' }.each do |k, v|
              @@logger.info "  IPv6: #{k}"      
              @coordinator_instance.write_data("#{@fq_node_id}.nics.#{iface_id}.ipaddress_v6", k)
              if iface_id == default_iface_ipv6
                @coordinator_instance.write_data("#{@fq_node_id}.ipaddress_v6", k)
              end
            end
          end
        end
      end
    
      
      # This hook is called after compilation and around the convergence phase.
      def register_converge_hook(chef_node)
        Aspect.new(:around, :calls_to => :converge, :for_types => [ Chef::Runner ]) do |jp, obj, *args|
          
          # We have to expand the run_list again to get a run list expansion trace that is needed
          # for intercepting recipies and associating them to roles. 
          # ASSUMPTION: that calling expand! again is stateless 
          # (it is also done in the Chef's client.rb but that is before our interception can kick in)
          if Chef::Config[:solo]
            @expansion = obj.run_context.node.expand!('disk')
          else
            @expansion = obj.run_context.node.expand!('server')
          end
          @run_list_trace = @expansion.run_list_trace
          @loaded_recipes = obj.run_context.loaded_recipes # array of loaded recipes
          Chef::Log.debug "Run List trace: #{@run_list_trace}"
          
          # process all write dependencies and make sure all default values are in the 'write_queue'          
          @write_dependencies.each do |dep| 
              @@logger.debug "Processing write dependency: #{dep}"
              params = dep["parameter"].split('.')
              node_value = chef_node
              # node_value.each { |k,v| @@logger.debug "node[#{k}] = #{v}" }
                @@logger.debug "searching for value using keys: #{params.inspect}"
              params.each { |p| node_value = node_value[p] unless node_value.nil? } # get the value of a parameter
              unless node_value.nil?
                write_loc = [dep["operation"], dep["parameter"].sub('.', '_')].compact.join('.')
                @@logger.debug %Q{Setting @write_queue[#{write_loc}] = "#{node_value}"}
                @write_queue[write_loc] = node_value
              end # unless node_value.nil?
              @@logger.debug "end of unless"
          end          
          @@logger.info "Processing 'queued writes' to coordinator..."
          @write_queue.each do |k,v|
            # filter out hostname and ipaddress - we don't want to write them again
            # the are written separately in the update_node_properties method
            unless IGNORED_WRITES.include?(k)
              @@logger.info %Q{  Writing queued item: #{k} = "#{v}"}
              @coordinator_instance.write_data("#{@fq_node_id}.#{k}", v)
            end
          end
          
          begin
            Timeout::timeout(@timeout) {
              jp.proceed
            }  # Timeout::timeout
          rescue Timeout::Error => te
            @@logger.error "Node configuration timed out"
            write_role_status(@current_role, Weaver::Coordination::CONFIG_STATUS::FAILED)
            # Rely on Chef framework and call back in NodeConfigStatus to handle setting the node status
            raise te
          rescue Exception => e
            @@logger.error "Exception in role #{@current_role}" #; trace is #{e.backtrace.join("\n")}"
            write_role_status(@current_role, Weaver::Coordination::CONFIG_STATUS::FAILED)
            # Rely on Chef framework and call back in NodeConfigStatus to handle setting the node status
            raise e
              # MK - work in progress to fix timeout error reporting
#          rescue Exception => e
#            @@logger.error "Exception in role #{@current_role}"
#            # Rely on Chef framework and call back in NodeConfigStatus to handle setting the node status
#            # Special handling for timeout so we can log it as the cause of the exception
#            is_timeout = e.backtrace.find {|bktrc| /Timeout::Error/ =~ bktrc }
#            unless is_timeout.nil?
#              raise CoordinationError.new("Node configuration timed out", e)
#            else 
#              raise CoordinationError.new("Re-raise configuration exception", e)
#            end 
          end
          
          # The last resource of the last recipe has been processed. Convergence is finished.
          Chef::Log.info "END ROLE: '#{@current_role}'" 
          write_role_status(@current_role, Weaver::Coordination::CONFIG_STATUS::CONFIGURED)
        end # Aspect.new(:after, ...)
      end        

      # The read/write hooks follow this idea:
      # We have two dependency arrays:
      #   - read_dependencies (i.e., instead of reading locally, read from coordinator)
      #   - write_dependenceis (i.e., instead of writing locally, also write value to coordinator)
      # Both arrays contain the parameter of the node in the coordinator, for example
      #   [ 'db', 'hostname' ]
      #
      # When a value in Chef is read, for example:  myval = node['db']['hostname']
      # it will trigger the this advice twice (or more depending on the chained []) with different args
      # values that will be chained together and then matched against the dependency:
      # ['db', 'hostname'].join('.') matches regex /x.y$/ (Note we match the parameter at the end of the dependency via the $ character.)
      # When a value in chef is written: node['db']['hostname'] is twill trigger as well with the
      # method_name == :[]= indicating that it was written.
      #
      # Depending on read and write, we select match a different dependency and trigger the read or write action.
      

      # Registers an attribute hook to intercept write calls to the 
      # Chef node data structure based on a set of dependencies given to the 
      # interceptor. 
      def register_write_attribute_hook()
        # Accumulate parameters to multi-level hashes e.g. node.set['dmgr']['hostname'] = 'foo'
        
        # In Chef 11+, node attributes that are written are wrapped in a Chef::Node::VividMash.
        # Experimentation shows that an aspect defined on :[]= will trigger only on the write itself.
        # That is, if an aspect is defined as:
        #    Aspect.new(:around, :calls_to => :[]=, :for_types => xx) do |jp, obj, *args|
        # And you have node.set['dmgr']['hostname'] = 'foo', the intercept will trigger only on the 
        # last layer of the Hash; in this example, 'hostname'.
        # This code supports only Chef 11+
        #
        # It is necessary to accumulate all of the parameters to multi-level hashes so that a 
        # mapping between Chef and weaver can be made.
        # Experimentation shows that the parameters for the other levels can be caught by 
        # defining aspects on the read of a VividMash. 
        # Further, this does not interfere with read interrupts (defined below in register_read_attribute_hook())
        # In the read case, triggers occur only on ImmutableHash objects. This is true even for 
        # a read reference to the same object being written.  For example:
        #   Chef::Log.info node['dgmr']['hostname']   <-- ImmutableMash
        #   node.set['dgmr']['hostname'] = 'foo'      <-- VividMash
        #   Chef::Log.info node['dgmr']['hostname']   <-- ImmutableMash
        
        # If we see a read on a VividHash, we are accumulating the parameters for a write
        # Clear the read accumulator just as a safety measure; it shouldn't be necessary.
        @arguments_read.clear
        Aspect.new(:around, :calls_to => :[], 
                   :for_types => [ Chef::Node::VividMash ]) do |jp, obj, *args|
          @@logger.debug "[read] [VividMash] Intercepting method #{jp.method_name} args: #{args.inspect}; cached arguments: #{@arguments_write}"
          # Ignore if it is already present at the end. In the read case (ImmutableMash),  
          # we noticed that reads are triggered both for :x and "x" in all cases.   
          # Not entirely sure it is needed for VividMash.
          unless !@arguments_write.empty? and @arguments_write.last == args[0].to_s
            # Add it our accumulation as long as the previous entry isn't the same
            @arguments_write << args[0].to_s # save the argument - because they are like node['x']['y']...
          end # unless !@arguments_write.empty? and @arguments_write.last == args[0].to_s
          
          # We don't need to worry about clearing this.  All reads of VividHash seem to be followed
          # by a single write.  Clear on the write below.
          jp.proceed
        end # Aspect.new(:around, :calls_to => :[]=

        Aspect.new(:around, :calls_to => :[]=,
                  :for_types => [ Chef::Node::VividMash ]) do |jp, obj, *args|
          @@logger.debug "[write] [VividMash] Intercepting method #{jp.method_name} args: #{args.inspect}; cached arguments: #{@arguments_write}"

          # Interception of a write on a VividMash indicates we have accumulated all of parameters 
          # (via the read interception above) except for the last one -- the current one).
          @arguments_write << args[0].to_s
          
          # look for a match to the set of write_dependencies
          param = @arguments_write.join('.')
          # @@logger.debug "[write] looking for /#{param}/"
          # Unlike read (below), we can do an exact match now -- we know we have all the arguments already
          matches = @write_dependencies.select { |d| d['parameter'] == param}
                       
          if matches.size == 1
            @@logger.debug "[write] Matched write dependency"
            match = matches[0]
            parameter = match['parameter'].gsub('.', '_')
            @@logger.info %Q{Write (immediate) to coordinator #{[@topology_id, @node_id, "#{match['operation']}.#{parameter}"].join('/')} = #{args[1].inspect}}
            @coordinator_instance.write_data("#{@fq_node_id}.#{match['operation']}.#{parameter}", args[1])
          elsif matches.size > 1 
            # This is an unexpected condition
            @@logger.warning "Couldn't write to coordinator, multiple matches #{matches}"
          else 
            # Didn't match a write dependency
            # @@logger.debug "[write] No matches"
          end # if match.size == 1
          @arguments_write.clear
          jp.proceed # no match in dependencies just proceed with usual result and clear saved arguments
        end # Aspect.new(:around, :calls_to => :[]=
                    
      end # def register_write_attribute_hook

      # Registers an attribute hook to intercept read calls to the 
      # Chef node data structure based on a set of dependencies given to the 
      # interceptor. 
      def register_read_attribute_hook()
        
        # In Chef 11+, node attributes that are read are wrapped in a Chef::Node::ImmutableMash.
        # Experimentation shows that an aspect defined on :[] will trigger on each parameter.
        # Furthermore, on first parameter, the trigger will fire twice, once for :x and once for "x".
        # It will also trigger for type Chef::Node::Attribute.  However, subsequent paramters
        # will fire only on ImmutableMash
        # This code supports only Chef 11+

        # It is necessary to accumulate all of the parameters to multi-level hashes so that a 
        # mapping between Chef and weaver can be made.

        Aspect.new(:around, :calls_to => :[], 
                   :for_types => [ Chef::Node::ImmutableMash ]) do |jp, obj, *args|
          @@logger.debug "[read] [ImmutableMash] Intercepting method #{jp.method_name} args: #{args.inspect}; cached arguments: #{@arguments_read}"

          # Ignore if it is already present at the end. In the read case (ImmutableMash),  
          # we noticed that reads are triggered both for :x and "x" in all cases.   
          if !@arguments_read.empty? and @arguments_read.last == args[0].to_s
            # @@logger.debug "[read] Ignoring #{args[0].inspect}; already in #{@arguments_read}"
            result = jp.proceed
          else
            # Accumulate the parameters                       
            @arguments_read << args[0].to_s
            # not necessary, but why not?
            @arguments_write.clear

            param = @arguments_read.join('.')
            # @@logger.debug "[read] looking for /#{param}/"
            match = @read_dependencies.select { |d| d['parameter'] =~ %r{#{param}} }
                         
            if match.size == 1 && match.first['parameter'] == param
              @@logger.debug "[read] dependency matching for argument #{args.inspect}; match = #{match.inspect}; match size = #{match.size}"
              late_binding_expr = match.first['late_binding_expr']
              late_binding_expr.gsub!(/__NEWLINE__/, "\n")
              @@logger.debug "  late_binding_expr is: '#{late_binding_expr}'"
              @@logger.debug "  values are: #{match.first['values']}"
              @@logger.debug "  attribute mapping is: #{match.first['mapping']}"
              lb = Weaver::Coordination::VmLateBinding.new(late_binding_expr, match.first['values'], match.first['mapping'])
              value = lb.eval(@coordinator_instance)
              @@logger.debug "Evaluated value: #{value}"
              result = value               
              @arguments_read.clear # clear out cached arguments            
            elsif match.size > 0
              # @@logger.debug "[read] At least one match"
              result = jp.proceed # match does not fully match @arguments_read
            else 
              # @@logger.debug "[read] No matches"
              @arguments_read.clear  
              result = jp.proceed # no match in dependencies just proceed with usual result and clear saved arguments
            end # if match.size == 1 && match.first['parameter'] == param
          end # @arguments.empty? and @arguments.last == args[0].to_s      
          result
        end # Aspect.new(:around, :calls_to => :[],
      end # def register_read_attribute_hook

      #
      # Intercepts the execution of a recipe. Before and after we check whether this recipe belongs
      # to the a given role and if it is the first or last one, so we can set the status accordingly.
      #
      def register_role_interception_hook()
        Aquarium::Aspects::Aspect.new(:around, :calls_to => :run_action, :for_types => [ Chef::Runner ]) do |jp, obj, *args|
          cookbook_name = args[0].cookbook_name
          recipe_name = args[0].recipe_name
          resource_name = args[0].resource_name
          next_recipe = "recipe[#{cookbook_name}::#{recipe_name}]"
          next_role = find_top_level_role(cookbook_name, recipe_name)
          
          if @current_role != next_role
            if @current_role != nil
              # Chef::Log.debug "current_role = #{@current_role}; next_role = #{next_role}"
              Chef::Log.info "END ROLE: '#{@current_role}'" 
              write_role_status(@current_role, Weaver::Coordination::CONFIG_STATUS::CONFIGURED)
            end
                        
            Chef::Log.info "BEGIN ROLE: '#{next_role}'"
            Chef::Log.info "role_dependencies: #{@role_dependencies}"
            Chef::Log.info "   #{sanitize_role_name(next_role)} --> #{@role_dependencies[sanitize_role_name(next_role)]}"
            unless @role_dependencies[sanitize_role_name(next_role)].nil?
              @role_dependencies[sanitize_role_name(next_role)].each do |dependency_id|
                Chef::Log.info "Wait for completion of '#{dependency_id}'"
                status = @coordinator_instance.read_data("#{dependency_id}.#{Weaver::Coordination::CONFIG_STATUS_STR}") do |st|
                  st =~ %r{#{Weaver::Coordination::CONFIG_STATUS::CONFIGURED}} or st =~ %r{#{Weaver::Coordination::CONFIG_STATUS::FAILED}}
                end # status = ...
                if status =~ %r{#{Weaver::Coordination::CONFIG_STATUS::FAILED}}
                  raise Weaver::Coordination::CoordinationError.new("Prerequisite action failed: #{dependency_id}")
                end # if status =~ ... FAILED
              end # @role_dependencies.each do 
            end # unless @role_dependencies[next_role].nil?
            write_role_status(next_role, Weaver::Coordination::CONFIG_STATUS::CONFIGURING)
            @current_role = next_role
          end

          if @current_recipe != next_recipe
            @current_recipe = next_recipe
          end

          begin
            jp.proceed
          rescue Exception => e
            Chef::Log.info "ROLE FAILED: '#{@current_role}'"
            write_role_status(@current_role, Weaver::Coordination::CONFIG_STATUS::FAILED)
            raise e
          end

          # check for end of last role: see register_converge_hook()
        end
      end
      
      private   
      
      # Writes the 'status' of a given role to Zookeeper 
      # * *Args*    :
      #   +role+  the name of role 
      #   +status+  the completion status of that roles (see Weaver::Coordination::CONFIG_STATUS)
      def write_role_status(role, status)
        @coordinator_instance.write_data(
          "#{@fq_node_id}.#{sanitize_role_name(role)}.#{Weaver::Coordination::CONFIG_STATUS_STR}",
          status)
      end

      # Converts a role name from 'role[foo]' to foo     
      # * *Args*    :
      #   +role_name+  the name of role in the format 'role[name_of_role]'
      # * +Returns+ :  the sanitized role name with the role[] stripped off
      def sanitize_role_name(role_name) 
        # we do not want "role[foo]" we just want "foo" so we return the first "match group"
        role_name.match(/role\[(.*)\]/).captures[0] unless role_name.nil?
      end

      # Fixed a bug in Chef's run_list_trace (it contains either the name of the role or the filename but not consistent)
      def fix_role_name(roles, role_name)      
       val = roles.select { |k| role_name == @expansion.fetch_role(k, nil).to_s }       
        retval = nil
        unless val.empty?
          retval = "role[#{val[0]}]"
        end        
        retval                 
      end
      
      # Apply the fix as described in the fix_role_name method if necessary
      def determine_and_fix_role_if_necessary(role)
        new_role = fix_role_name(@expansion.roles, role)
        if new_role
          Chef::Log.debug "Replacing role: #{new_role}"
          role = new_role
        end
        role
      end
      
      # Finds the top level role that a given recipe belongs to
      # * *Args*   :
      #   +run_list+ the full run_list trace that Chef stores in the run_context
      #   +cookbook_name+ the name of the cookbook that is being executed
      #   +recipe_name+ the name of the recipe that is being executed
      def find_top_level_role(cookbook_name, recipe_name) 
        # we need to look for both: <cookbook_name>::<recipe_name> and <cookbook_name> (with the default recipe)
        fq_recipe = "#{cookbook_name}::#{recipe_name}" 
        fq_default_recipe = "#{cookbook_name}::default"
        
        Chef::Log.debug "Trace to search is: #{@run_list_trace.inspect}"
        role_map = @run_list_trace.select { |k,v| v.include?("recipe[#{fq_recipe}]") || v.include?("recipe[#{cookbook_name}]") || v.include?("recipe[#{fq_default_recipe}]") }
        Chef::Log.debug "Found role in trace: #{role_map}"
        role = nil
        if role_map.size == 1 
          role = determine_and_fix_role_if_necessary(role_map.keys[0])
          while !@run_list_trace['top level'].include? role # we walk up the run_list_trace
            role_map = @run_list_trace.select { |k,v| v.include? role }               
            if role_map.nil? || role_map.empty?
              Chef::Log.error "[Weaver Chef Interception] Cannot find role #{role} in run_list_trace: #{@run_list_trace}"
              Chef::Log.error "Exiting..."
              exit 1
            else
              role = determine_and_fix_role_if_necessary(role_map.keys[0])
            end
          end
        elsif @loaded_recipes.include?(fq_recipe) || @loaded_recipes.include?(fq_default_recipe) # this case handles "include_recipe" statements that don't show up in the runlist
          role = @current_role # we assume they belong to the current role 
        else
          # this should actually not happen (unless there is a bug in the chef expansion)
          Chef::Log.error "[Weaver Chef Interception] Cannot find mapping between recipe '#{cookbook_name}::#{recipe_name}' to its top level role"    
          exit 1
        end
        role 
      end
      
      def convert_dependency(dependency) 
        retval = {}
        ['node', 'operation', 'parameter', 'late_binding_expr'].each do |key| 
          retval[key] = dependency[key] unless dependency[key].nil?
        end # ['node', 'operation', 'parameter', 'late_binding_expr'].each do
        ['values','mapping'].each do |key|
          unless dependency[key].nil?
            copy = {}
            dependency[key].each { |k,v| copy[k] = v }
            retval[key] = copy
          end # unless dependency[key].nil?
        end # ['values','mapping'].each do
        return retval
      end # def convert_read_dependency
      
      def log_dependency(dependency)
        return dependency.inspect
      end # def log_dependency
      
    end
  end
end