require 'zk'
require 'singleton'
require 'weaver/coordination/logging'

# Singleton service to support creation of ZKMock objects (to mock ZK)
# Maintains the single global storage used by the ZKMock objects
class ZKMockService
  include Singleton
  
  attr_accessor :global_store
  
  def initialize()
    @global_store = {}
  end
  
  # Create ZKMock object; if the connection to the servers doesn't already exist, create an entry in the global table
  def create (servers)
    raise RuntimeError, "error connecting to zookeeper: 22" if servers.nil?
    # TODO validate that id is a list of servers:port
    # TODO break servers into list and have id for each all pointing to the same object
    @global_store[servers] = { "/" => ["",{}], "/zookeeper" => ["", {}] } if @global_store[servers].nil?
    return ZKMock.new(servers)
  end
  
  # Express the intent that for the given mocked ZK servers, a particlar value should be returned for a given key (path)
  def expect (servers, path, value)
    zk = ZKMock.new(servers)
    zk.mkdir_p(path)
    zk.set(path, value)
  end

  # how do I indicate that only ZKMock should call these?
  def get(id, path)
    store = get_store(id) # throws exception if not available
    return store[path]
  end
  
  # set a value in the global store
  def set(id, path, value)
    store = get_store(id) # throws exception if not available
    store[path] = value
  end
  
  # select elements from global store baesd on evaluation using provided proc
  def select(id, &block)
    store = get_store(id) # throws exception if not available
    return store.select(&block) unless store.nil?
    return []
  end
  
  # delete an element from the global store 
  def delete(id, path)
    store = get_store(id) # throws exception if not available
    store.delete(path) unless store.nil?
  end
  
  private 
  def get_store (id)
    store = @global_store[id]
    raise Zookeeper::Exceptions::NotConnected, "connection state is closed" if store.nil?
    return store
  end
  
end # class ZKMockService

# Mock ZK object
class ZKMock
  
  def initialize(servers)
    @connected = true
    @id = servers
  end
  
  def connected?
    return @connected
  end
  
  def connecting?
    return false
  end
  
  def closed?
    return !connected?
  end
  
  def reopen  
    retval = @connected ? :connected : :connecting
    @connected = true
    return retval
  end
  
  def close!
    @connected = false
  end
  
  def close
    @connected = false
  end
  
  # check if entry is present
  def exists?(path, options = {})
    raise Zookeeper::Exceptions::NotConnected unless @connected
    return !ZKMockService.instance.get(@id, path).nil?
    # return !@store[path].nil?
  end
  
  # create an element; the parent needs to already exist
  # can optionally set the value 
  def create (path, data="")
    raise Zookeeper::Exceptions::NotConnected unless @connected
    raise ZK::Exceptions::NoNode, %Q{inputs: {:path=>"#{path}, :data=>"#{data}"}} if !exists?(File.dirname(path))
    raise ZK::Exceptions::NodeExists, %Q{{:path=>"#{path}", :data=>"#{data}", :ephemeral=>false, :sequence=>false}} if exists?(path)
    set_internal(path,data)
    return path
  end
  
  # get entry for given path; returns a two element array: [value, metadata]
  def get (path)
    raise Zookeeper::Exceptions::NotConnected unless @connected
    raise ZK::Exceptions::NoNode, %Q{inputs: "#{path}"} if !exists?(path)
    return ZKMockService.instance.get(@id, path)
    # return @store[path]
  end
  
  # set a value for a given path; the element must already exist
  def set (path, data)
    raise Zookeeper::Exceptions::NotConnected unless @connected
    raise ZK::Exceptions::NoNode, %Q{inputs: {:path=>"#{path}, :data=>"#{data}"}} if !exists?(path)
    set_internal(path,data)
  end
  
  # identify the names of the children of an element
  def children(path)
    raise Zookeeper::Exceptions::NotConnected unless @connected
    raise ZK::Exceptions::NoNode, %Q{inputs: {:path=>"#{path}"}} if !exists?(path)
    ZKMockService.instance.select(@id) { |key,value| !(%r{^#{path}/?[^/]*$} =~ key).nil? and !(path == key)}.map{ |key, value| key }.map { |path| path.split('/').last }
    # @store.select { |key,value| !(%r{^#{path}} =~ key).nil? and !(path == key)}.map{ |key, value| key }
  end

  # create an element including all parents if needed
  def mkdir_p (path)
    raise Zookeeper::Exceptions::NotConnected unless @connected
    retval = nil
    create_path = ""
    components = path.split('/')
    until components.empty? 
      create_path << (%r{/$} !~ create_path ? "/#{components.shift}" : components.shift)
      retval = exists?(create_path) ? nil : create_path
      create_internal(create_path) unless exists?(create_path)
    end
    return retval
  end
  
  # remove an element and all its children (recursively)
  def rm_rf (path)
    raise Zookeeper::Exceptions::NotConnected unless @connected
    children(path).each { |child| rm_rf("#{path}/#{child}") }
    ZKMockService.instance.delete(@id, path)
    return [path]
  end
  
  # register an event handler
  def register(path, &block) 
    raise Zookeeper::Exceptions::NotConnected unless @connected
    return MockEventHandler.new
  end
  
  private
  def set_internal (path, data = "")
    ZKMockService.instance.set(@id, path, [data.to_s, {}])
    # @store[path] = [data.to_s, {}]
  end
  
  alias :create_internal :set_internal
  
end # class ZKMock

# Mock ZK EventHandler.  Doesn't do anything
class MockEventHandler
  include Weaver::Coordination::Logging

  def initialize
    @@logger.debug "MockEventHandler.new() called; trace = #{caller}"
    # not implemented
  end
  def unsubscribe
    # not implemented
  end
end # class EventHandler

# Mock ZK Watcher. Doesn't do anything
class MockWatcher
  include Weaver::Coordination::Logging
  
  def initialize
    @@logger.debug "MockWatcher.new() called; trace = #{caller}"
    # not implemented
  end
  def completed?
    return true
  end
end # class MockWatcher

 