require 'weaver/core/logging'
require 'weaver/core/late_binding'

module Weaver
  # \Weaver::Service provides services for developers to operate on \Weaver models.
  #
  # The only service defined at this time is the PropagationChecker.
  module Service
    
    # Author:: Ed Snible
    module PropagationChecker
      include Weaver::Core

      # These are the attributes that change on startup
      private
      KEY_ATTRIBUTE_IDS = [:ip_address, :hostname]
        
      # Return a Hash of Script to Hash of String/paramname to Set of Attribute
      def self.model_propagations(env)
        retval = {}
        env.flatten_children.select { |child| child.respond_to?(:automation) }.each do |entity|
          entity.automation.each { |id, script| retval[script] = model_script_propagations(script) }
        end
        return retval
      end
      
      def self.model_script_propagations(script)
        retval = {}
        script.amd.values.each do |attr|         
          retval[attr.id] = attr.unresolved_value.attribute_dependencies(script.scope) if attr.unresolved_value.kind_of?(LateBinding)
        end
        return retval
      end

      # Returns a Hash of Attribute to Array of Script
      # 'propagations' must be a Hash of Script to Hash of String/paramname to Set of Attribute
      # such as one returned by model_propagations()
      def self.invert_propagations(propagations)
        retval = Hash.new { |h, k| h[k] = [] }
        propagations.each { |script, hash| hash.each { |paramname, attrs| attrs.each { |attr, mapping|
          retval[attr].push(script) 
        } } }
        return retval
      end
      
      # Returns a Hash of Attribute to Array of Node
      # 'propagations' must be a Hash of Script to Hash of String/paramname to Set of Attribute
      # such as one returned by model_propagations()
      def self.node_dependencies(propagations)
        retval = Hash.new { |h, k| h[k] = [] }
        propagations.each { |script, hash| hash.each { |paramname, attrs| attrs.each { |attr, mapping|
          retval[attr].push(script.node) 
        } } }
        return retval
      end
      
      # If hostname or ip_address is nil, make sure changing them only affects later scripts 
      def self.valid_startup_order(env, startup_order = env.startup_order)
        
        # Get all nodes dependent on any attribute changes
        deps = node_dependencies(model_propagations(env))
        return check_startup_order(deps, startup_order)
      end
      
      def self.generate_startup_order(env)
        
        # Get all nodes dependent on any attribute changes
        deps = node_dependencies(model_propagations(env))
        
        # Note This is a trivial algorithm that tries all permutations of nodes having effects on other nodes.
        # A better algorithm could exploit parallel startup and would scale better.  Anyone improving
        # this code must also fix the Environment.validate_collect() to handle suggestions including parallelism.

        # Divide into nodes whose ip_address affects scripts and nodes that don't
        nodes_with_script_effects = Set.new
        infrastructure_nodes = env.flatten_children.select { |entity| entity.kind_of?(Node) && entity.realized_by.nil? }
        infrastructure_nodes.each do |node|
          nodes_dependent_on_this_nodes_ipaddress = Set.new()
          KEY_ATTRIBUTE_IDS.each do |id|
            attr = node.amd[id]
            if attr.value.nil? && deps[attr].size > 0
#              @@logger.debug "found a node whose IP address effects a script, it's attribute[:#{node.id}][:#{id}]"
              nodes_with_script_effects.add(node)
            end
          end
        end

        @@logger.debug "generate_startup_order, found nodes #{nodes_to_ids(nodes_with_script_effects.to_a)} have script effects"
        
        safe_to_start_last = (Set.new(infrastructure_nodes) - nodes_with_script_effects).to_a
        @@logger.debug "generate_startup_order, found nodes #{nodes_to_ids(safe_to_start_last)} don't affect subsequent node scripts"
        
        nodes_with_script_effects.to_a.permutation do |permutation|
          candidate_order = permutation.concat(safe_to_start_last)
          if check_startup_order(deps, candidate_order)
            return nodes_to_ids(candidate_order)
          end
        end
        
        return nil
          
      end
      
      private
      def self.nodes_to_ids(nodes)
        return nodes.map { |node| node.id }
      end
      
      private
      def self.check_startup_order(deps, startup_order)
        # track of the initial startup_order sequence that can start without error
        nodes_already_started = Set.new()
        
        # Check to see if starting each node effects script parameters on nodes already started
        startup_order.each do |node_or_array|
          group_of_nodes = node_or_array.kind_of?(Array) ? node_or_array : [node_or_array]
          nodes_dependent_on_these_nodes_ipaddress = Set.new()
          group_of_nodes.each do |node|
            KEY_ATTRIBUTE_IDS.each do |id|
              attr = node.amd[id]
              nodes_dependent_on_these_nodes_ipaddress.merge(deps[attr]) if attr.value.nil?
            end
          end
 
          # I have commented this out because 'nagios.new_client.ip_address' makes every node
          # dependent upon itself but I don't understand why a node can be dependent on itself
          # but we can't have cycles of dependencies.  TODO Restore or come to an understanding         
#          group_of_nodes.each do |node|
#            return false if nodes_dependent_on_these_nodes_ipaddress.member?(node) # Node is dependent upon itself/it's group!
#          end
          nodes_dependent_on_these_nodes_ipaddress.each { |dependent_node| return false if nodes_already_started.member?(dependent_node) }
          nodes_already_started.merge(group_of_nodes)
        end
        
        return true
      end
    end
      
  end
end
